/*
 * clnt_udpa.c		Support for async RPC over UDP. This code is
 *			a modified version of clnt_udp.c from Sun's
 *			RPC 4.0 code. The original license is reproduced
 *			below.
 *
 * All modifications for async RPC by Olaf Kirch, <okir@monad.swb.de>
 */
/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)clnt_udp.c 1.39 87/08/11 Copyr 1984 Sun Micro";
#endif

/*
 * clnt_udp.c, Implements a UDP/IP based, client side RPC.
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <stdio.h>
#include <rpc/rpc.h>
#include <sys/socket.h>
#include <linux/net.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <errno.h>
#include <rpc/pmap_clnt.h>
#include <signal.h>
#include <fcntl.h>
#include "misc.h"

extern int errno;

/*
 * UDP bases client side rpc operations
 */
static enum clnt_stat	clntudpa_call();
static void		clntudpa_abort();
static void		clntudpa_geterr();
static bool_t		clntudpa_freeres();
static bool_t           clntudpa_control();
static void		clntudpa_destroy();
static void		clntudpa_sigio();
static void		clntudpa_initialize();
static struct cua_req * clntudpa_createreq(CLIENT *cl, u_long proc, 
				xdrproc_t xargs, caddr_t argsp,
				xdrproc_t xresults, caddr_t resultp,
				struct timeval utimeout, int cookie);
static void		clntudpa_queuereq(struct cua_req *cr);
static enum clnt_stat	clntudpa_marshall(struct cua_req *cr);
static enum clnt_stat	clntudpa_send(struct cua_req *cr);
static void		clntudpa_sigio(int sig);
static void		clntudpa_timeout(int sig, void *clnt_data);
static void		clntudpa_recv(struct cua_req *cr, int inlen);
static void		clntudpa_lock_queue(void);
static void		clntudpa_release_queue(void);
static void		clntudpa_freereq(struct cua_req *cr);

static struct clnt_ops udp_ops = {
	clntudpa_call,
	clntudpa_abort,
	clntudpa_geterr,
	clntudpa_freeres,
	clntudpa_destroy,
	clntudpa_control
};

/* 
 * Private data kept per client handle
 */
struct cua_data {
	int		   cua_sock;
	bool_t		   cua_closeit;
	struct sockaddr_in cua_raddr;
	int		   cua_rlen;
	u_long		   cua_program;
	u_long		   cua_version;
	struct timeval	   cua_wait;
	struct timeval     cua_total;
	struct rpc_err	   cua_error;
	u_int		   cua_sendsz;
	u_int		   cua_recvsz;
	u_short		   cua_xid;
	void		   (*cua_callback)();
	/* struct cua_req     *cua_requests; */
};

struct cua_req {
	struct cua_req	   *cua_nextreq;	/* next queued request */
	CLIENT		   *cua_client;		/* back pointer */
	struct cua_data    *cua_data;		/* back pointer */
	u_long		   cua_proc;		/* procedure # */
	u_long		   cua_xid;		/* current XID */
	int		   cua_cookie;		/* call id for req matching */

	XDR		   cua_outxdrs;
	xdrproc_t	   cua_xdrarg;		/* args XDR function */
	u_int		   cua_xdrpos;		/* XDR posn after header */
	caddr_t		   cua_argp;		/* user-provided arguments */
	xdrproc_t	   cua_xdrres;		/* result XDR function */
	caddr_t		   cua_resp;		/* user-provided res buffer */
	char *		   cua_outbuf;		/* complete RPC datagram */
	int		   cua_outlen;
	char *		   cua_inbuf;		/* RPC receive buffer */
	int		   cua_inlen;

	int		   cua_timer;		/* xtimer id */
	struct timeval	   cua_total;		/* total timeout */
	struct timeval	   cua_wait;		/* time between resends */
	struct timeval	   cua_waited;		/* already waited */
};

static CLIENT *		udpa_client[FD_SETSIZE];
static struct cua_req *	udpa_requests = NULL;
static fd_set		udpa_fdset;
static int		udpa_initialized = 0;
static int		udpa_qlock = 0;
static sigset_t		udpa_sigset;


/*
 * Create a UDP based client handle.
 * If *sockp<0, *sockp is set to a newly created UPD socket.
 * If raddr->sin_port is 0 a binder on the remote machine
 * is consulted for the correct port number.
 * NB: It is the clients responsibility to close *sockp.
 * NB: The rpch->cl_auth is initialized to null authentication.
 *     Caller may wish to set this something more useful.
 *
 * wait is the amount of time used between retransmitting a call if
 * no response has been heard;  retransmition occurs until the actual
 * rpc call times out.
 *
 * sendsz and recvsz are the maximum allowable packet sizes that can be
 * sent and received.
 */
CLIENT *
clntudpa_bufcreate(raddr, program, version, wait, sockp, sendsz, recvsz, callb)
	struct sockaddr_in *raddr;
	u_long program;
	u_long version;
	struct timeval wait;
	register int *sockp;
	u_int sendsz;
	u_int recvsz;
	void (*callb)();
{
	CLIENT *cl;
	register struct cua_data *cu;
	struct timeval now;
	int	flags;

	cl = (CLIENT *)mem_alloc(sizeof(CLIENT));
	if (cl == NULL) {
		(void) fprintf(stderr, "clntudpa_create: out of memory\n");
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		goto fooy;
	}

	cu = (struct cua_data *)mem_alloc(sizeof(*cu));
	if (cu == NULL) {
		(void) fprintf(stderr, "clntudpa_create: out of memory\n");
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		goto fooy;
	}

	(void)gettimeofday(&now, (struct timezone *)0);
	if (raddr->sin_port == 0) {
		u_short port;
		if ((port =
		    pmap_getport(raddr, program, version, IPPROTO_UDP)) == 0) {
			goto fooy;
		}
		raddr->sin_port = htons(port);
	}
	cl->cl_ops = &udp_ops;
	cl->cl_private = (caddr_t)cu;
	cu->cua_program = program;
	cu->cua_version = version;
	cu->cua_raddr = *raddr;
	cu->cua_rlen = sizeof (cu->cua_raddr);
	cu->cua_wait = wait;
	cu->cua_total.tv_sec = -1;
	cu->cua_total.tv_usec = -1;
	cu->cua_sendsz = ((sendsz + 3) / 4) * 4;
	cu->cua_recvsz = ((recvsz + 3) / 4) * 4;
	cu->cua_callback = callb;
	cu->cua_xid = getpid() ^ now.tv_sec ^ now.tv_usec;

	if (*sockp < 0) {
		int dontblock = 1;

		*sockp = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (*sockp < 0) {
			rpc_createerr.cf_stat = RPC_SYSTEMERROR;
			rpc_createerr.cf_error.re_errno = errno;
			goto fooy;
		}
		/* attempt to bind to prov port */
		(void)bindresvport(*sockp, (struct sockaddr_in *)0);
		/* the sockets rpc controls are non-blocking */
		(void)ioctl(*sockp, FIONBIO, (char *) &dontblock);
		cu->cua_closeit = TRUE;
	} else {
		cu->cua_closeit = FALSE;
	}

	if (*sockp >= FD_SETSIZE) {
		if (cu->cua_closeit)
			close(*sockp);
		goto fooy;
	}
	clntudpa_initialize();

	/* Set up async IO on socket */
	flags = fcntl(*sockp, F_GETFL) | FASYNC;
	if (fcntl (*sockp, F_SETFL, flags) < 0) {
		fprintf(stderr, "can't set FASYNC on UDP socket\n");
	}
	if (fcntl (*sockp, F_SETOWN, getpid()) < 0) {
		fprintf(stderr, "can't set owner pid on socket\n");
	}

	FD_SET(*sockp, &udpa_fdset);
	udpa_client[*sockp] = cl;

	cu->cua_sock = *sockp;
	cl->cl_auth = authnone_create();
	return (cl);
fooy:
	if (cu)
		mem_free((caddr_t)cu, sizeof(*cu));
	if (cl)
		mem_free((caddr_t)cl, sizeof(CLIENT));
	return ((CLIENT *)NULL);
}

CLIENT *
clntudpa_create(raddr, program, version, wait, sockp, callb)
	struct sockaddr_in *raddr;
	u_long program;
	u_long version;
	struct timeval wait;
	register int *sockp;
	void (*callb)();
{

	return(clntudpa_bufcreate(raddr, program, version, wait, sockp,
	    UDPMSGSIZE, UDPMSGSIZE, callb));
}

static enum clnt_stat 
clntudpa_call(cl, proc, xargs, argsp, xresults, resultsp, utimeout, cookie)
	register CLIENT	*cl;		/* client handle */
	u_long		proc;		/* procedure number */
	xdrproc_t	xargs;		/* xdr routine for args */
	caddr_t		argsp;		/* pointer to args */
	xdrproc_t	xresults;	/* xdr routine for results */
	caddr_t		resultsp;	/* pointer to results */
	struct timeval	utimeout;	/* seconds to wait before giving up */
	int		cookie;		/* magic cookie from caller */
{
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;
	register struct cua_req *cr;

	cr = clntudpa_createreq(cl, proc, xargs, argsp, xresults, resultsp,
					utimeout, cookie);
	if (cr == NULL) {
		return (cu->cua_error.re_status = RPC_CANTENCODEARGS);
	}

	if (clntudpa_marshall(cr) != RPC_SUCCESS) {
		clntudpa_freereq(cr);
		return (cu->cua_error.re_status);
	}

	if (clntudpa_send(cr) != RPC_SUCCESS) {
		clntudpa_freereq(cr);
		return (cu->cua_error.re_status);
	}

	clntudpa_queuereq(cr);

	/*
	 * Hack to provide rpc-based message passing
	 * Doesn't work currently.
	 */
	/*
	if (timeout.tv_sec == 0 && timeout.tv_usec == 0) {
		return (cu->cua_error.re_status = RPC_TIMEDOUT);
	}
	 */

	/* We've sent the request. The rest is done by someone else */
	return (cu->cua_error.re_status);
}

/*
 * Create an RPC request and fill in the RPC packet header.
 */
static struct cua_req *
clntudpa_createreq(cl, proc, xargs, argsp, xresults, resultsp, utimeout, cookie)
	register CLIENT	*cl;		/* client handle */
	u_long		proc;		/* procedure number */
	xdrproc_t	xargs;		/* xdr routine for args */
	caddr_t		argsp;		/* pointer to args */
	xdrproc_t	xresults;	/* xdr routine for results */
	caddr_t		resultsp;	/* pointer to results */
	struct timeval	utimeout;	/* seconds to wait before giving up */
	int		cookie;		/* user-supplied request id */
{
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;
	register struct cua_req *cr;
	struct rpc_msg call_msg;

	cr = (struct cua_req *) mem_alloc(sizeof(struct cua_req));
	if (cr == NULL) {
		return NULL;
	}

	cr->cua_outbuf = mem_alloc(cu->cua_sendsz);
	cr->cua_inbuf = mem_alloc(cu->cua_sendsz);
	if (cr->cua_outbuf == NULL || cr->cua_inbuf == NULL) {
		goto fooy;
	}

	call_msg.rm_xid = cr->cua_xid = cu->cua_xid++;
	call_msg.rm_direction = CALL;
	call_msg.rm_call.cb_rpcvers = RPC_MSG_VERSION;
	call_msg.rm_call.cb_prog = cu->cua_program;
	call_msg.rm_call.cb_vers = cu->cua_version;
	xdrmem_create(&(cr->cua_outxdrs), cr->cua_outbuf,
			cu->cua_sendsz, XDR_ENCODE);
	if (! xdr_callhdr(&(cr->cua_outxdrs), &call_msg)) {
		goto fooy;
	}
	cr->cua_xdrpos = XDR_GETPOS(&(cr->cua_outxdrs));

	cr->cua_client = cl;
	cr->cua_cookie = cookie;
	cr->cua_data = cu;
	cr->cua_proc = proc;
	cr->cua_xdrarg = xargs;
	cr->cua_xdrres = xresults;
	cr->cua_argp = argsp;
	cr->cua_resp = resultsp;
	cr->cua_wait = cu->cua_wait;

	if (cu->cua_total.tv_usec == -1) {
		cr->cua_total = utimeout;	/* use supplied timeout */
	} else {
		cr->cua_total = cu->cua_total;	/* use default timeout */
	}
	cr->cua_waited.tv_sec  = 0;
	cr->cua_waited.tv_usec = 0;

	return cr;

fooy:
	if (cr && cr->cua_outbuf)
		mem_free(cr->cua_outbuf, cu->cua_sendsz);
	if (cr && cr->cua_inbuf)
		mem_free(cr->cua_inbuf, cu->cua_recvsz);
	if (cr)
		mem_free (cr, sizeof(struct cua_req));
	return NULL;
}

/*
 * Queue an RPC request.
 */
static void
clntudpa_queuereq(cr)
	struct cua_req *cr;
{
	clntudpa_lock_queue();

	cr->cua_timer = xtimer_run(cr->cua_wait, clntudpa_timeout, NULL);
	cr->cua_nextreq = udpa_requests;
	udpa_requests = cr;

	clntudpa_release_queue();
}

/*
 * Serialize the RPC data.
 */
static enum clnt_stat
clntudpa_marshall(cr)
	struct cua_req *cr;
{
	register struct cua_data *cu = cr->cua_data;
	CLIENT	*cl = cr->cua_client;
	register XDR *xdrs;

	xdrs = &(cr->cua_outxdrs);
	xdrs->x_op = XDR_ENCODE;
	XDR_SETPOS(xdrs, cr->cua_xdrpos);

	/* Bump the XID */
	*(u_long *)(cr->cua_outbuf) = cr->cua_xid = cu->cua_xid++;

	if ((! XDR_PUTLONG(xdrs, (long *)&(cr->cua_proc))) ||
	    (! AUTH_MARSHALL(cl->cl_auth, xdrs)) ||
	    (! (*cr->cua_xdrarg)(xdrs, cr->cua_argp)))
		return (cu->cua_error.re_status = RPC_CANTENCODEARGS);
	cr->cua_outlen = (int)XDR_GETPOS(xdrs);

	return (cu->cua_error.re_status = RPC_SUCCESS);
}

/*
 * This is the actual transmit routine.
 */
static enum clnt_stat
clntudpa_send(cr)
	struct cua_req *cr;
{
	struct cua_data	*cu = cr->cua_data;

	if (sendto(cu->cua_sock, cr->cua_outbuf, cr->cua_outlen, 0,
	    (struct sockaddr *)&(cu->cua_raddr), cu->cua_rlen)
	    != cr->cua_outlen) {
		cu->cua_error.re_errno = errno;
		return (cu->cua_error.re_status = RPC_CANTSEND);
	}

	return (cu->cua_error.re_status = RPC_SUCCESS);
}

/*
 * A reply with matching XID has been received for a particular request.
 * Invoke the callback.
 */
static void
clntudpa_recv(cr, inlen)
	struct cua_req	*cr;
	int		inlen;
{
	struct cua_data 	*cu = cr->cua_data;
	CLIENT			*cl = cr->cua_client;
	struct rpc_msg		reply_msg;
	XDR			reply_xdrs;
	/* int			nrefreshes = 2; */
					/* number of times to refresh cred */
	bool_t			ok;


	/*
	 * decode and validate the response
	 */
	reply_msg.acpted_rply.ar_verf = _null_auth;
	reply_msg.acpted_rply.ar_results.where = cr->cua_resp;
	reply_msg.acpted_rply.ar_results.proc = cr->cua_xdrres;

	xdrmem_create(&reply_xdrs, cr->cua_inbuf, (u_int)inlen, XDR_DECODE);
	ok = xdr_replymsg(&reply_xdrs, &reply_msg);
	/* XDR_DESTROY(&reply_xdrs);  save a few cycles on noop destroy */
	if (ok) {
		_seterr_reply(&reply_msg, &(cu->cua_error));
		if (cu->cua_error.re_status == RPC_SUCCESS) {
			if (! AUTH_VALIDATE(cl->cl_auth,
				&reply_msg.acpted_rply.ar_verf)) {
				cu->cua_error.re_status = RPC_AUTHERROR;
				cu->cua_error.re_why = AUTH_INVALIDRESP;
			}
			if (reply_msg.acpted_rply.ar_verf.oa_base != NULL) {
				/* Questionable .. */
				cr->cua_outxdrs.x_op = XDR_FREE;
				(void)xdr_opaque_auth(&(cr->cua_outxdrs),
				    &(reply_msg.acpted_rply.ar_verf));
			} 
		}  /* end successful completion */
		else {
			/* maybe our credentials need to be refreshed ... */
			/*
			if (nrefreshes > 0 && AUTH_REFRESH(cl->cl_auth)) {
				nrefreshes--;
				clntudpa_queuereq(cr);
				clntudpa_marshall(cr);
				clntudpa_send(cr);
				return;
			}
			 */
		}  /* end of unsuccessful completion */
	}  /* end of valid reply message */
	else {
		cu->cua_error.re_status = RPC_CANTDECODERES;
	}

	/* Now invoke the callback function */
	if (cu->cua_callback != NULL) {
		(cu->cua_callback)(cl, cr->cua_proc, cr->cua_cookie);
	}

	/* ... and free the request */
	clntudpa_freereq(cr);

	return;
}

static void
clntudpa_freereq(cr)
	struct cua_req *cr;
{
	struct cua_data *cu = cr->cua_data;

	if (cr->cua_inbuf)
		mem_free(cr->cua_inbuf, cu->cua_recvsz);
	if (cr->cua_outbuf)
		mem_free(cr->cua_outbuf, cu->cua_recvsz);
	XDR_DESTROY(&(cr->cua_outxdrs));
}

/*
 * This is the SIGIO signal handler. It executes with all signals blocked,
 * so it doesn't need special protection.
 */
static void
clntudpa_sigio(int sig)
{
	struct timeval	wait = {0, 0};
	static int	lastsock = 0;
	fd_set		readfds;
	int		fd, last;
	struct cua_req	*cr, **crp;
	struct cua_data *cu;
	char *		buffer;
	u_long		xid;
	int		inlen;

	/* fprintf(stderr, "clntudpa_sigio called\n"); */
	memcpy(&readfds, &udpa_fdset, sizeof(fd_set));
	while (select(FD_SETSIZE, &readfds, NULL, NULL, &wait) < 0) {
		if (errno != EINTR)
			return;
	}

	/* Loop over all file descriptors and check if there's something
	 * in it for us.
	 */
	fd = last = lastsock;
	do {
		if (++fd >= FD_SETSIZE)
			fd = 0;
		if (udpa_client[fd] == NULL || !FD_ISSET(fd, &readfds))
			continue;

		/*
		 * Receive the datagram.
		 */
		/* clntudpa_lock_queue(); */
		crp = &udpa_requests;
		for (; (cr = *crp) != NULL; crp = &((*crp)->cua_nextreq)) {
			if ((*crp)->cua_client == udpa_client[fd])
				break;
		}
		if (cr == NULL) { /* No requests pending. Discard */
			int	foo;
			recvfrom(fd, &foo, sizeof(foo), 0, NULL, NULL);
			/* clntudpa_release_queue(); */
			continue;
		}
		/* Borrow buffer from first request no matter what. */
		buffer = cr->cua_inbuf;
		cu = cr->cua_data;

		do {
			inlen = recvfrom(fd, buffer, cu->cua_recvsz, 0,
						NULL, NULL);
		} while (inlen < 0 && errno == EINTR);
		if (inlen < sizeof(u_long)) {
			/* clntudpa_release_queue(); */
			continue;
		}
		xid = *(u_long *)buffer;

		for (; *crp != NULL; crp = &((*crp)->cua_nextreq)) {
			if ((*crp)->cua_client == udpa_client[fd] &&
			    (*crp)->cua_xid == xid)
				break;
		}
		if (*crp == NULL) {
			/* clntudpa_release_queue(); */
			continue;
		}
		if (*crp != cr) {
			cr->cua_inbuf = (*crp)->cua_inbuf;
			(*crp)->cua_inbuf = buffer;
			cr = *crp;
		}
		xtimer_stop(cr->cua_timer);
		*crp = cr->cua_nextreq;
		/* clntudpa_release_queue(); */

		/* We now assume we have a valid response. Process. */
		clntudpa_recv(cr, inlen);

		last = fd;
	} while (fd != lastsock);
	lastsock = last;
}

/*
 * A timeout occurred on an RPC request.
 */
static void
clntudpa_timeout(timerid, clnt_data)
	int timerid;
	void *clnt_data;
{
	struct cua_req	*cr, **crp;
	struct cua_data	*cu;

	/* This is for debugging purposes */
	clntudpa_lock_queue();
	clntudpa_sigio(0);
	clntudpa_release_queue();

	/* Lock the queue */
	clntudpa_lock_queue();

	/* Find the timed-out request */
	for (crp = &udpa_requests; *crp != NULL; crp = &((*crp)->cua_nextreq)) {
		if ((*crp)->cua_timer == timerid)
			break;
	}
	if ((cr = *crp) == NULL) {
		clntudpa_release_queue();
		return;	/* huh? */
	}

	/* Remove the timed-out request from the queue */
	*crp = cr->cua_nextreq;

	/* No need to lock the queue anymore. */
	clntudpa_release_queue();

	/* Reschedule or delete */
	timeval_bump(cr->cua_waited, cr->cua_wait);
	if (timeval_lesseq(cr->cua_waited, cr->cua_total)) {
		clntudpa_send(cr);
		clntudpa_queuereq(cr);
	} else {
		cu = cr->cua_data;
		if (cu->cua_callback != NULL) {
			cu->cua_error.re_status = RPC_TIMEDOUT;
			(cu->cua_callback)(cr->cua_client,
					cr->cua_proc, cr->cua_cookie);
		}
		clntudpa_freereq(cr);
	}
}

static void
clntudpa_lock_queue()
{
	sigset_t	block;

	sigfillset(&block);
	sigprocmask(SIG_BLOCK, &block, &udpa_sigset);

	if (udpa_qlock) {
		/* Aaargh! We called clntudpa_lock_queue with the queue
		 * already locked. */
		fprintf(stderr, "RPC UDPA queue locking failed!\n");
		return;
	}
	udpa_qlock = 1;
}

static void
clntudpa_release_queue()
{
	if (!udpa_qlock)
		return;

	udpa_qlock = 0;
	sigprocmask(SIG_SETMASK, &udpa_sigset, NULL);
}

static void
clntudpa_geterr(cl, errp)
	CLIENT *cl;
	struct rpc_err *errp;
{
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;

	*errp = cu->cua_error;
}


static bool_t
clntudpa_freeres(cl, xdr_res, res_ptr)
	CLIENT *cl;
	xdrproc_t xdr_res;
	caddr_t res_ptr;
{
	/*
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;
	register XDR *xdrs = &(cu->cua_outxdrs);

	xdrs->x_op = XDR_FREE;
	return ((*xdr_res)(xdrs, res_ptr));
	 */
	xdr_free(xdr_res, res_ptr);
	return 1;
}

static void 
clntudpa_abort(/*h*/)
	/*CLIENT *h;*/
{
}

static bool_t
clntudpa_control(cl, request, info)
	CLIENT *cl;
	int request;
	char *info;
{
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;

	switch (request) {
	case CLSET_TIMEOUT:
		cu->cua_total = *(struct timeval *)info;
		break;
	case CLGET_TIMEOUT:
		*(struct timeval *)info = cu->cua_total;
		break;
	case CLSET_RETRY_TIMEOUT:
		cu->cua_wait = *(struct timeval *)info;
		break;
	case CLGET_RETRY_TIMEOUT:
		*(struct timeval *)info = cu->cua_wait;
		break;
	case CLGET_SERVER_ADDR:
		*(struct sockaddr_in *)info = cu->cua_raddr;
		break;
	default:
		return (FALSE);
	}
	return (TRUE);
}
	
static void
clntudpa_destroy(cl)
	CLIENT *cl;
{
	register struct cua_data *cu = (struct cua_data *)cl->cl_private;

	if (cu->cua_closeit) {
		(void)close(cu->cua_sock);
	}
	mem_free((caddr_t)cu, sizeof(*cu));
	mem_free((caddr_t)cl, sizeof(CLIENT));

	/* We should check for requests that are still pending here and
	 * free them. Problem is, we don't know whether to call xdr_free
	 * on the argument and result pointers or not.
	 */
}

static void
clntudpa_initialize()
{
	struct sigaction	act;
	int			i;

	if (udpa_initialized)
		return;

	FD_ZERO(&udpa_fdset);
	for (i = 0; i < FD_SETSIZE; i++)
		udpa_client[i] = NULL;

	act.sa_handler = clntudpa_sigio;
	act.sa_flags = SA_RESTART;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGVTALRM);
	sigaddset(&act.sa_mask, SIGALRM);
	sigaddset(&act.sa_mask, SIGIO);
	sigaction(SIGIO, &act, NULL);

	udpa_initialized = 1;
}
