/*
 * file.c	File handle stuff. 
 *
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "unfsfh.h"
#include "nlm.h"

/*
 * List of local files.
 */
static lockd_file *		local_file_tbl;

/*
 * Locate file handle information. Currently, this is still based on
 * the FH information alone. Later, we should obtain the inode information
 * and use that as an index into the list.
 */
lockd_file *
file_lookup(lockd_host *host, netobj *fh, int local)
{
	lockd_file	*file_tbl;
	lockd_file	*p;
	struct stat	stb;

	if (local && fh->n_len != NFS_FHSIZE)
		return NULL;

	file_tbl = local? local_file_tbl : host->file_tbl;

	/* First, try to locate file by handle data. */
	for (p = file_tbl; p != NULL; p = p->nxt) {
		if (netobj_cmp(&(p->fh), fh))
			continue;
		if (!local)
			return p;
		if (fstat(p->fd, &stb) < 0 ||
		    stb.st_dev != p->i_dev || stb.st_ino != p->i_inum ||
		    file_getversion(p->fd) != p->i_version) {
			log(L_WARNING, "Found stale file handle");
			return NULL;
		}
		return p;
	}
	return file_create(host, fh, local);
}

/* 
 * Create new FH and insert into file list.
 * If lookup is non-zero, we try to open the file corresponding to the
 * handle.
 */
lockd_file *
file_create(lockd_host *host, netobj *fh, int local)
{
	lockd_file	**tblp;
	lockd_file	*file;

	tblp = local? &local_file_tbl : &host->file_tbl;

	if (local && fh->n_len != NFS_FHSIZE) {
		/* log(L_ERROR, "file handle size != NFS_FHSIZE"); */
		errno = EINVAL;
		return NULL;
	}
	file = (lockd_file *) xmalloc (sizeof(*file));

	file->nxt	    = NULL;
	file->prv	    = NULL;
	file->granted	    = NULL;
	file->blocked	    = NULL;
	file->acknowledged  = NULL;
	file->retry_blocked = 0;
	file->update_locks  = 0;
	file->read_only     = 0;
	netobj_copy(&(file->fh), fh);

	if (!local) {
		file->name = NULL;
		file->fd = -1;
	} else {
		struct stat	stb;

		file->name = unfsfh_buildpath((struct nfs_fh *) fh->n_bytes);
		if (file->name == NULL)
			goto stale;
		if ((file->fd = open(file->name, O_RDWR)) < 0) {
			if ((file->fd = open(file->name, O_RDONLY)) < 0)
				goto stale;
			file->read_only = 1;
		}
		if (fstat(file->fd, &stb) < 0)
			goto stale;
		file->i_dev = stb.st_dev;
		file->i_inum = stb.st_ino;
		file->i_version = file_getversion(file->fd);
	}

	file->nxt = *tblp;
	*tblp = file;

	return file;

stale:
	file_delete(host, file);
	errno = ESTALE;
	return NULL;
}

/*
 * Relase file when it's no longer used. Close fd and delete if
 * no more locks.
 */
void
file_release(lockd_host *host, lockd_file *file)
{
	if (file->granted || file->blocked || file->acknowledged)
		return;
	file_delete(host, file);
}

/*
 * Return file descriptor associated with file.
 */
int
file_fd(lockd_file *file)
{
	return file->fd;
}

char *
file_pathname(lockd_file *file)
{
	return file->name? file->name : "<remote>";
}

/*
 * Get the inode version. Return 0 for now, but use ex2tfs ioctl later
 * if applicable.
 */
int
file_getversion(int fd)
{
	return 0;
}

/*
 * Retry the list of blocked locks.
 */
void
file_retry_blocked(lockd_file *file)
{
	lockd_lock	*lock, **prevp;

	prevp = &(file->blocked);
	while ((lock = *prevp) != NULL) {
		if (lock_grant_clntreq(lock) == LCK_GRANTED)
			continue;
		prevp = &((*prevp)->file.nxt);
	}
	file->retry_blocked = 0;
}

/*
 * Delete a file handle
 */
void
file_delete(lockd_host *host, lockd_file * file)
{
	lockd_file	*file_tbl = host->file_tbl;
	lockd_file	*tmp;

	if ((tmp = file->prv) != NULL) {
		tmp->nxt = file->nxt;
	} else {
		file_tbl = file->nxt;
	}

	if ((tmp = file->nxt) != NULL) {
		tmp->prv = file->prv;
	}

	netobj_free(&(file->fh));
	if (file->name)
		xfree(file->name);
	if (file->fd >= 0)
		close(file->fd);

	xfree (file);
}

/*
 * Loop over all (local) files.
 */
void
file_foreach(void (*func)(lockd_file *))
{
	lockd_file	*file;

	for (file = local_file_tbl; file != NULL; file = file->nxt)
		func(file);
}

/*
 * Initialize file handle cache 
 */
void
file_init(void)
{
	local_file_tbl = NULL;
}

