/*
 *	clnt.c - lock owner information
 *
 *	orig
 *	260295	okir	added generic list insert/remove
 *
 */

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <netdb.h>

#include "nlm.h"

#define CLNT_MAXHASH		0x7F
#define CLNT_HASH(owner)	tolower(*(unsigned char *)(owner))

static lockd_host *		clients[CLNT_MAXHASH + 1];
static char			myhostname[LM_MAXSTRLEN+1];

/*
 * Locate a client in our cache. Later, the rqstp information
 * should be used for some sort of authentication. For calls
 * coming in via KLM, rqstp is passed as NULL.
 */
lockd_host *
host_lookup (struct svc_req *rqstp, char *owner, int authenticate)
{
	struct in_addr	addr;
	lockd_host	*p;

	if (rqstp != NULL)
		addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
	else
		addr.s_addr = htonl(INADDR_ANY);

	for (p = clients[CLNT_HASH (owner)]; p != NULL; p = p->nxt) {
		if (authenticate) {
			int	i;

			for (i = 0; i < p->nr_addrs; i++) {
				if (p->more_addrs[i].s_addr == addr.s_addr)
					return p;
			}
		} else {
			if (!strcasecmp (p->name, owner))
				return p;
		}
	}

	if (rqstp == NULL)
		return NULL;

	return host_create (rqstp, owner, authenticate);
}

/*
 * Create a client record
 */
lockd_host *
host_create (struct svc_req * rqstp, char *owner, int authenticate)
{
	lockd_host	*host, *tmp;
	struct hostent	*hp;
	struct in_addr	**ap, addr;
	int		i, hash;

	host = (lockd_host *) xmalloc (sizeof (lockd_host));

	if ((hp = gethostbyname(owner)) == NULL) {
		log(L_ERROR, "Can't get address of %s", owner);
		xfree (host);
		return NULL;
	}
	if (hp->h_addrtype != AF_INET) {
		log(L_ERROR, "address type of %s is not AF_INET",
					owner);
		xfree (host);
		return NULL;
	}
	ap = (struct in_addr **) hp->h_addr_list;

	/*
	 * Verify the originator of the call. This is a very poor type
	 * of authentication.
	 */
	if (rqstp != NULL && authenticate) {
		addr = svc_getcaller (rqstp->rq_xprt)->sin_addr;
		for (i = 0; ap[i] != NULL; i++) {
			if (addr.s_addr == ap[i]->s_addr)
				break;
		}
		if (ap[i] == NULL) {
			log(L_ERROR, "hostname %s doesn't match address",
					owner);
			xfree (host);
			return NULL;
		}
		host->addr = addr;
	} else {
		host->addr = *(ap[0]);
	}

	/* Now copy all addresses */
	for (i = 0; ap[i] != NULL; i++) ;
	host->more_addrs = (struct in_addr *) xmalloc(i * sizeof(addr));
	host->nr_addrs = i;
	while (--i >= 0) {
		host->more_addrs[i] = *(ap[i]);
	}

	host->name = xstrdup (owner);

	host->r_granted = NULL;
	host->r_blocked = NULL;
	host->l_granted = NULL;
	host->l_blocked = NULL;
	host->l_waiting = NULL;
	host->l_reclaim = NULL;
	host->use_msg   = 0;
	host->state	= 0;
	host->monitored = 0;

	host->file_tbl  = NULL;

	host->prv = NULL;
	hash = CLNT_HASH (owner);
	if ((tmp = clients[hash]) != NULL) {
		tmp->prv = host;
	}
	host->nxt = tmp;
	clients[hash] = host;

	return host;
}

/*
 * Delete a client record
 */
void
host_delete (lockd_host * host)
{
	lockd_host *	tmp;

	/*
	 * Make sure there are no outstanding lock requests.
	 */
	if (host->r_granted != NULL || host->r_blocked != NULL ||
	    host->l_granted != NULL || host->l_blocked != NULL ||
	    host->l_waiting != NULL || host->l_reclaim != NULL) {
		log (L_ERROR, "deleted client %s still holds locks!",
				host->name);
	}

	if (host->file_tbl != NULL) {
		log(L_ERROR, "deleted host %s still has files", host->name);
	}

	/*
	 * Unlink client from the doubly linked list
	 */
	if ((tmp = host->prv) != NULL) {
		tmp->nxt = host->nxt;
	} else {
		clients[CLNT_HASH (host->name)] = host->nxt;
	}

	if ((tmp = host->nxt) != NULL) {
		tmp->prv = host->prv;
	}

	/*
	 * ... and reclaim the memory
	 */
	xfree (host->more_addrs);
	xfree (host->name);
	xfree (host);
}

char *
host_myname(void)
{
	return myhostname;
}

/*
 * Initialize client cache
 */
void
host_init (void)
{
	int i;

	for (i = 0; i < CLNT_MAXHASH; i++) {
		clients[i] = NULL;
	}

	/*
	 * Get our own hostname
	 */
	if (gethostname(myhostname, LM_MAXSTRLEN) < 0) {
		log(L_ERROR, "can't get my hostname");
		strcpy(myhostname, "foobar");
	} else {
		struct hostent * hp;

		myhostname[LM_MAXSTRLEN] = '\0';
		if ((hp = gethostbyname(myhostname)) == NULL) {
			log(L_ERROR, "can't resolve my own hostname");
		} else {
			strncpy(myhostname, hp->h_name, LM_MAXSTRLEN);
			myhostname[LM_MAXSTRLEN] = '\0';
		}
	}
	log(L_DEBUG, "my hostname is \"%s\"", myhostname);
}
