/*
 * lklm.c	Service routines for LKLM.
 *
 * Copyright (C) 1995, O. Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include "lock.h"
#include "lklm.h"
#include "nlm.h"
#include "log.h"

static lockd_req * lklm_makereq(lockd_host *, lockd_proto *, int,
				struct lklm_lock *, int, int);
static int	lklm_test_svc(struct lklm_request *, lockd_proto *);
static int	lklm_lock_svc(struct lklm_request *, lockd_proto *);
static int	lklm_unlock_svc(struct lklm_request *, lockd_proto *);
static int	lklm_cancel_svc(struct lklm_request *, lockd_proto *);
static int	lklm_test_status(int status, lockd_req *rq);
static int	lklm_lock_status(int status, lockd_req *rq);
static int	lklm_unlock_status(int status, lockd_req *rq);
static int	lklm_cancel_status(int status, lockd_req *rq);

typedef int	(*lklm_fn_t)(struct lklm_request *, lockd_proto *);
static lklm_fn_t	lklm_fntable[] = {
	lklm_test_svc,
	lklm_lock_svc,
	lklm_unlock_svc,
	lklm_cancel_svc,
};
#define LKLM_CALL_MAX	(sizeof(lklm_fntable)/sizeof(lklm_fntable[0]))

/*
 * How much to backoff if server is in grace period.
 */
static struct timeval	clnt_grace_backoff = { 5, 0 };
int			lklmfd = -1;

void
lklm_register(void)
{
#ifdef DEBUG
	struct sockaddr_un	sun;

	bzero(&sun, sizeof(sun));
	sun.sun_family = AF_UNIX;
	strcpy(sun.sun_path, _PATH_LKLMSOCK);

	unlink(_PATH_LKLMSOCK);
	if ((lklmfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		log(L_FATAL, "can't create lklm socket: %s", strerror(errno));
	if (bind(lklmfd, (struct sockaddr *) &sun, sizeof(sun)) < 0)
		log(L_FATAL, "can't bind to lklm socket: %s", strerror(errno));
	if (listen(lklmfd, 5) < 0)
		log(L_FATAL, "can't listen on lklm socket: %s",
						strerror(errno));
#else
	log(L_FATAL, "lklm_init: only debug mode supported for now");
#endif
}

int
lklm_dispatch(void)
{
	struct lklm_request	rqst;
	lockd_proto		*prot;
	int			len, call;

	if ((len = recv(lklmfd, &rqst, sizeof(rqst), 0)) < 0) {
		log(L_WARNING, "lklm recv: %s", strerror(errno));
		return 0;
	}
	if (len < sizeof(rqst)) {
		log(L_WARNING,
			"lklm recv: short read (incompatible protocol?)");
		return 0;
	}
	if (rqst.prot_version != LKLM_VERSION) {
		log(L_WARNING,
			"lklm dispatch: incompatible LKLM protocol version");
		return 0;
	}

	call = rqst.request;
	if (call < 0 || call > LKLM_CALL_MAX) {
		log(L_WARNING, "lklm dispatch: invalid request");
		return 0;
	}

	if ((prot = proto_find(rqst.lock_proto)) == NULL)
		return 0;

	return lklm_fntable[call](&rqst, prot);
}

static void
lklm_reply(int pid, int status, struct lklm_holder *holder)
{
	struct lklm_reply	rply = { LKLM_VERSION, pid, status, };

	if (holder)
		memcpy(&rply.u.holder, holder, sizeof(*holder));

	if (send(lklmfd, &rply, sizeof(rply), 0) < 0) {
		log(L_WARNING, "lklm sendreply failed: %s", strerror(errno));
	}
}

/*
 * Test for the presence of a lock. Maybe we should check if the target
 * host refers to ourself and take a short cut.
 */
static int
lklm_test_svc(struct lklm_request *rqstp, lockd_proto *prot)
{
	struct lklm_testarg	*argp = &rqstp->u.test;
	struct lklm_lock	*alock = &argp->alock;
	lockd_host		*host;
	lockd_req		*rq;
	int			ex = argp->exclusive;

	if ((host = host_lookup(NULL, argp->alock.server_name, 0)) == NULL) {
		lklm_reply(alock->pid, LKLM_GRANTED, NULL);
		return 1;
	}

	/* For duplicate requests, ignore the call. */
	if (req_find(host->l_waiting, RQ_TEST, alock, ex, 0)) {
		return 0;
	}

	/*
	 * Send the request and insert it into host->l_waiting. If proto
	 * cannont send the request, the callback is invoked which
	 * returns an error to the kernel.
	 */
	rq = lklm_makereq(host, prot, RQ_TEST, alock, ex, 0);
	req_queue(rq, lklm_test_status);

	return 0;
}

/*
 * Status callback for a test request. 
 */
static int
lklm_test_status(int status, lockd_req *rq)
{
	struct lklm_holder	holder;

	switch (status) {
	case RQ_DENIED:
		if (rq->proto->getconflock(rq, &holder) != 0)
			lklm_reply(rq->pid, status, &holder);
		else
			lklm_reply(rq->pid, RQ_FAILED, NULL);
		break;
	case RQ_DENIED_GRACE_PERIOD:
		rq->wait = clnt_grace_backoff;
		req_retry(rq);
		return 0;
	default:
		lklm_reply(rq->pid, status, NULL);
		break;
	}
	return 1;
}

/*
 * Create a lock on the server host.
 */
static int
lklm_lock_svc(struct lklm_request *rqstp, lockd_proto *prot)
{
	struct lklm_lockarg *	argp = &rqstp->u.lock;
	struct lklm_lock *	alock = &argp->alock;
	lockd_host *		host;
	lockd_req *		rq;
	int			ex = argp->exclusive;
	int			bl = argp->block;

	if ((host = host_lookup(NULL, argp->alock.server_name, 0)) == NULL) {
		lklm_reply(alock->pid, LKLM_NOHOST, NULL);
		return 0;
	}

	/* Ignore duplicate requests. */
	if (req_find(host->l_waiting, RQ_LOCK, alock, ex, bl) ||
	    req_find(host->l_reclaim, RQ_LOCK, alock, ex, bl) ||
	    req_find(host->l_blocked, RQ_LOCK, alock, ex, bl)) {
		return 0;
	}

	/* Forward the request to the NLM server. */
	rq = lklm_makereq(host, prot, RQ_LOCK, alock, ex, bl);
	req_queue(rq, lklm_lock_status);

	return 0;
}

/* 
 * Status callback for lock request.
 */
static int
lklm_lock_status(int status, lockd_req *rq)
{
	int	retval = 1;

	if (rq->reclaim) {
		if (status != RQ_GRANTED) {
			/* Bad things happening. Complain and send lock
			 * holder a SIGLOST.
			 */
			log(L_WARNING, "failed to reclaim lock, status %d",
					status);
			kill(SIGLOST, rq->pid);
		}
		return 1;
	}
	if (status == RQ_TIMEOUT) {
		lklm_reply(rq->pid, status, NULL);
		return 1;
	}

	switch (status) {
	case RQ_BLOCKED:
		req_insert(&(rq->host->l_blocked), rq);
		status = LKLM_WORKING;
		retval = 0;
		break;
	case RQ_DENIED_GRACE_PERIOD:
		rq->wait = clnt_grace_backoff;
		req_retry(rq);
		return 0; /* don't reply to client, retry later */
	}

	lklm_reply(rq->pid, status, NULL);
	return 0;
}

/*
 * Cancel a blocked lock on the server host.
 */
static int
lklm_cancel_svc(struct lklm_request *rqstp, lockd_proto *prot)
{
	struct lklm_lockarg	*argp = &rqstp->u.lock;
	struct lklm_lock *	alock = &(argp->alock);
	lockd_host *		host;
	lockd_req *		rq;
	int			ex = argp->exclusive;
	int			bl = argp->block;

	if ((host = host_lookup(NULL, argp->alock.server_name, 0)) == NULL) {
		lklm_reply(alock->pid, LKLM_NOHOST, NULL);
		return 0;
	}

	/* Check for request retransmissions */
	if (req_find(host->l_waiting, RQ_CANCEL, alock, ex, bl) ||
	    req_find(host->l_waiting, RQ_UNLOCK, alock, ex, bl)) {
		return 0;
	}

	/* This part here is a little tricky. The following races can
	 * occur:
	 *  a.	The lock request is still in flight. Bad things happen
	 *	if the cancel request arrives at the server before the
	 *	lock request.
	 *
	 *  b.	The lock request has been granted already.
	 *
	 * We solve a) by ignoring it (for the moment). To solve b)
	 * we should check if we currently hold this lock and send out
	 * an UNLOCK if we do. However, for the time being, it's easier
	 * to just send the UNLOCK no matter what.
	 */

	/* a: discard any pending LOCK requests. Not optimal. */
	if ((rq = req_find(host->l_waiting, RQ_LOCK, alock, ex, bl)) != NULL) {
		req_unlink(rq);
		req_abort(rq);
		req_free(rq);
	}

	/* b: send an UNLOCK request */
	rq = lklm_makereq(host, prot, RQ_UNLOCK, alock, ex, bl);
	req_queue(rq, lklm_unlock_status);

	/* Check if there's a blocked lock request at all. If there is none,
	 * it may be that we already processed the cancel, but our reply to
	 * the client got lost. Just tell it that all is well.
	 */
	if ((rq = req_find(host->l_blocked, RQ_LOCK, alock, ex, bl)) == NULL) {
		lklm_reply(alock->pid, LKLM_GRANTED, NULL);
		return 1;
	}

	req_unlink(rq);
	req_free(rq);

	/* Forward the request to the NLM server. */
	rq = lklm_makereq(host, prot, RQ_CANCEL, alock, ex, bl);
	req_queue(rq, lklm_cancel_status);

	return 0;
}

/*
 * RPC status for an asynchronous NLM_CANCEL_MSG call to the server.
 */
static int
lklm_cancel_status(int status, lockd_req *rq)
{
	switch (status) {
	case RQ_DENIED_GRACE_PERIOD:
		rq->wait = clnt_grace_backoff;
		req_retry(rq);
		return 0;
	default:
		lklm_reply(rq->pid, status, NULL);
	}
	return 1;
}

/*
 * Free a lock on the server host.
 */
static int
lklm_unlock_svc(struct lklm_request *rqstp, lockd_proto *prot)
{
	struct lklm_unlockarg *	argp = &rqstp->u.unlock;
	struct lklm_lock *	alock = &argp->alock;
	lockd_host *		host;
	lockd_req *		rq;

	if ((host = host_lookup(NULL, argp->alock.server_name, 0)) == NULL) {
		lklm_reply(alock->pid, LKLM_NOHOST, NULL);
		return 0;
	}

	/*
	 * If there's a matching blocked request, just remove it. 
	 */
	if ((rq = req_find(host->l_blocked, RQ_LOCK, alock, 1, -1))) {
		log(L_WARNING, "weird: unlock request for blocked lock");
		req_unlink(rq);
		req_free(rq);

		lklm_reply(alock->pid, LKLM_GRANTED, NULL);
		return 1;
	}

	/* Don't send out duplicate requests. */
	if (req_find(host->l_waiting, RQ_UNLOCK, alock, -1, -1)) {
		log(L_DEBUG, "ignored duplicate KLM_UNLOCK request");
		return 0;
	}

	/* Forward the request to the NLM server. */
	rq = lklm_makereq(host, prot, RQ_UNLOCK, alock, 0, 0);
	req_queue(rq, lklm_unlock_status);

	return 0;
}

/*
 * Status of an NLM_UNLOCK_MSG call to the server.
 */
static int
lklm_unlock_status(int status, lockd_req *rq)
{
	switch (status) {
	case RQ_DENIED_GRACE_PERIOD:
		rq->wait = clnt_grace_backoff;
		req_retry(rq);
		return 0;
	default:
		lklm_reply(rq->pid, status, NULL);
	}
	return 1;
}

static lockd_req *
lklm_makereq(lockd_host *host, lockd_proto *prot, int type,
				struct lklm_lock *lkreq, int ex, int bl)
{
	lockd_req	*rq;

	rq = req_alloc(host, prot, type, lkreq->pid);

	netobj_set(&rq->fh, lkreq->fh.data, lkreq->fh.len);
	rq->l_offset	= lkreq->l_offset;
	rq->l_len	= lkreq->l_len;
	rq->exclusive	= ex;
	rq->block	= bl;

	return rq;
}
