/*
 *	lock.h		Lock related definition and procedures.
 *
 */

#ifndef LOCKD_LOCK_H
#define LOCKD_LOCK_H

#include "nlm_prot.h"
/* #include "klm_prot.h" */
#include "lklm.h"

/*
 * Forward declarations.
 */
typedef struct lockd_lock lockd_lock;
typedef struct lockd_host lockd_host;
typedef struct lockd_file lockd_file;
typedef struct lockd_req  lockd_req;

#include "proto.h"

/*
 * The main lock structure. This is used on the server for both Unix locks
 * and DOS shares.
 */
struct lockd_lock {
    unsigned short	mode;		/* The normal lock types use this:
					 * exclusive:	fsm_DRW (= 3)
					 * else:	fsm_DN  (= 0)
					 * DOS shares are stored as-is.
					 */
    unsigned short	access;		/* The normal lock types use this:
					 * exclusive:	fsa_RW (= 3)
					 * else:	fsa_R  (= 1)
					 * DOS shares are stored as-is.
					 */
    unsigned char	exclusive;	/* 1: exclusive lock (DOS W/RW shares
					 * count as exclusive, too). This
					 * is somewhat redundant but makes
					 * the code more readable.
					 */
    unsigned char	async;		/* 1: lock was requested  with
					 * NLM_LOCK_MSG
					 */
    unsigned char	block;		/* 1: process requested blocking.
					 * Used only on the client host.
					 */

    unsigned long	l_start;	/* start & length of lock */
    unsigned long	l_end;
    unsigned long	l_length;	/* 0 for lock until EOF */
    netobj		oh;		/* owner handle */
    u_int		pid;		/* pid of lock holder */

    struct {
        lockd_lock *	nxt;		/* links on per-file lock list */
        lockd_lock *	prv;
	lockd_file *	back;		/* back pointer */
	int		linked;
    } file;
    struct {
        lockd_lock *	nxt;		/* links on per-host lock list */
        lockd_lock *	prv;
	lockd_host *	back;		/* back pointer */
	int		linked;
    } host;
};

#define LCKACCS_EXCL	fsa_RW
#define LCKACCS_SHARED	fsa_R
#define LCKMODE_EXCL	fsm_DRW
#define LCKMODE_SHARED	fsm_DN

/*
 * This defines the callback function for forwarded lock requests.
 * It is invoked as soon as the status of the RPC call to the server 
 * is known. status describes whether the RPC call was successful; rq is
 * the request. The return value of this function determines if the request
 * should be freed (1) or not (0).
 *
 * This function is invoked with all signals turned off.
 */
typedef int		(*rqstatus_fn_t)(int status, lockd_req *rq);

/*
 * This structure holds all interesting data about a forwarded lock 
 * request.
 * Whatever you do, make sure this struct is not deleted before
 * any associated RPC calls have been dequeued.
 */
struct lockd_req {
	lockd_proto *	proto;		/* lock protocol */
	lockd_host *	host;		/* server host */
	lockd_req *	next;		/* doubly linked list */
	lockd_req *	prev;
	lockd_req *	nextreq;	/* list of in-flight requests */

	int		type;		/* kind of request (LOCK, UNLOCK etc) */
	int		pid;

	unsigned long	l_offset;	/* lock information */
	unsigned long	l_len;
	netobj		oh;		/* may be void for non-NLM */
	netobj		fh;
	unsigned char	exclusive;
	unsigned char	block;

	unsigned char	reclaim;	/* 1 for reclaiming */
	unsigned char	wasgranted;	/* 1 if lock was granted before */

	void *		result;		/* Data returned by server */
	rqstatus_fn_t	callback;	/* status callback */
	int		cookie;		/* request cookie */
	int		state;		/* state of request */

	struct timeval	wait;		/* Re-xmit requests during reclaim */

	int		timerid;	/* Used by both wait and retry */
	int		retries;
};

/* Lock request states. */
enum {
	LCKSTATE_INITIAL = 0,
	LCKSTATE_REQSENT,
	LCKSTATE_REPLYRECVD,
	LCKSTATE_PROCESSED
};

/* Request types */
enum {
	RQ_TEST = 0,
	RQ_LOCK,
	RQ_UNLOCK,
	RQ_CANCEL,
};

/* Request results. These largely follow the NLM4 reqturn codes */
enum {
	RQ_TIMEOUT = -1,
	RQ_GRANTED,
	RQ_DENIED,
	RQ_DENIED_NOLOCKS,
	RQ_BLOCKED,
	RQ_DENIED_GRACE_PERIOD,
	RQ_DEADLCK,
	RQ_ROFS,
	RQ_STALE_FH,
	RQ_FBIG,

	RQ_QUEUED = 256,
	RQ_FAILED,
};

/*
 * Public functions for lock manipulation.
 */
lockd_lock *	lock_testlock(lockd_host *clnt, lockd_file *file, 
					nlm_lock *lkreq, int exclusive);
int		lock_createlock(lockd_host *clnt, lockd_file *file, 
					nlm_lock *lkreq, int exclusive);
int		lock_unlock(lockd_host *clnt, lockd_file *file, 
					nlm_lock *lkreq);
int		lock_createshare(lockd_host *host, lockd_file *file,
					nlm_share *shreq);
int		lock_unshare(lockd_host *host, lockd_file *file,
					nlm_share *shreq);
void		lock_unlock_all(lockd_host *lock);
void		lock_freeall(lockd_lock **head);
void		lock_block_clntreq(lockd_host *clnt, lockd_file *file,
					nlm_lock *lkreq, int exclusive,
					int async);
int		lock_grant_clntreq(lockd_lock *lock);
void		lock_grant_status(int status, lockd_lock *lock);
void		lock_cancel_clntreq(lockd_host *clnt, lockd_file *file,
					nlm_lock *lkreq, int exclusive);
void		lock_dump(void);

lockd_req *	req_alloc(lockd_host *host, lockd_proto *prot,
					int type, int pid);
int		req_queue(lockd_req *rq, rqstatus_fn_t callback);
/*
int		req_queue(lockd_host *host, struct svc_req *rqstp,
					rqstatus_fn_t callback, int type,
					klm_lock * lkreq, int exclusive,
					int block);
 */
void		req_free(lockd_req *rq);
void		req_abort(lockd_req *rq);
/* void		req_process_result(lockd_req *rq, int status, void *resultp); */
void		req_process_result(lockd_req *rq, int status);
void		req_status(int status, lockd_req *req);
void		req_retry(lockd_req *rq);
int		req_grant_blocked(lockd_host *host, nlm_testargs *argp);
lockd_req *	req_find(lockd_req *head, int type, struct lklm_lock *lkreq,
					int block, int exclusive);
lockd_req *	req_match(int cookie);
void		req_insert(lockd_req **head, lockd_req *rq);
void		req_unlink(lockd_req *rq);
void		req_lock_queue(int lockit);

/*
 * Private functions. These must be used only from lock_s.c and lock_c.c
 */
lockd_lock	*lk_alloc(lockd_file *file, lockd_host *clnt);
void		lk_init(lockd_lock *lock, lockd_host *host, lockd_file *file,
					unsigned long offset,
					unsigned long len, netobj *oh,
					int pid, int exclusive);
void		lk_quick_init(lockd_lock *lock, lockd_host *host,
					lockd_file *file,
					unsigned long offset,
					unsigned long len, netobj *oh,
					int pid, int exclusive);
void		lk_free(lockd_lock *lock);
void		lk_insert_file(lockd_lock **head, lockd_lock *prev,
					lockd_lock *lock);
void		lk_append_file(lockd_lock **head, lockd_lock *lock);
void		lk_insert_host(lockd_lock **head, lockd_lock *prev,
					lockd_lock *lock);
void		lk_unlink_host(lockd_lock **head, lockd_lock *lock);
void		lk_unlink_file(lockd_lock **head, lockd_lock *lock);
lockd_lock *	lk_check(lockd_lock *new, lockd_lock ***prevp);
int		lk_lock(int action, lockd_lock *new);
void		lk_update(lockd_file *file, unsigned long start,
					unsigned long end, int action);
lockd_lock *	lk_find(lockd_lock *head, nlm_lock *lkreq,
					int exclusive);
int		lk_compare(lockd_lock *lock, nlm_lock *req, int excl);
void		lk_dumpfile(lockd_file *file);
void		lk_dumplock(lockd_lock *lock);
void		unlock_region(lockd_file *file, unsigned long start,
					unsigned long len);
void		lock_region(lockd_file *file, unsigned long start,
					unsigned long len, int excl);

/*
 * Action types for lk_lock.
 */
enum {
	AC_LOCK,
	AC_RELOCK,
	AC_PRELOCK,
	AC_UNLOCK
};

#endif /* LOCKD_LOCK_H */
