/*
 * lock_p.c	Private lock-management functions.
 *
 */

#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include "nlm.h"

/*
 * Create a new lock.
 */
lockd_lock *
lk_alloc (lockd_file *file, lockd_host *clnt)
{
	lockd_lock *	lk;

	lk = (lockd_lock *) xmalloc(sizeof(lockd_lock));

	lk->file.back = file;
	lk->host.back = clnt;

	lk->host.prv = lk->host.nxt = NULL;
	lk->file.prv = lk->file.nxt = NULL;

	lk->host.linked = 0;
	lk->file.linked = 0;

	return lk;
}

/*
 * Duplicate an existing lock.
 */
lockd_lock *
lk_dup (lockd_lock *lock)
{
	lockd_lock	*new;

	new = lk_alloc(lock->file.back, lock->host.back);
	new->mode      = lock->mode;
	new->access    = lock->access;
	new->exclusive = lock->exclusive;
	new->l_start   = lock->l_start;
	new->l_length  = lock->l_length;
	new->l_end     = lock->l_end;
	new->pid       = lock->pid;
	netobj_copy(&(new->oh), &(lock->oh));

	return new;
}

/*
 * Initialize a lock given a NLM lock request. This is the quick version,
 * because we forego copying the owner handle.
 */
void
lk_quick_init (lockd_lock *lock, lockd_host *host, lockd_file *file,
		u_long offset, u_long len, netobj *oh, int pid, int exclusive)
{
	memset(lock, 0, sizeof(*lock));
	lock->file.back = file;
	lock->host.back = host;
	lock->mode      = exclusive? LCKMODE_EXCL : LCKMODE_SHARED;
	lock->access    = exclusive? LCKACCS_EXCL : LCKACCS_SHARED;
	lock->exclusive = exclusive;
	lock->l_start   = offset;
	lock->l_length  = len;
	lock->l_end     = len? offset + len : ULONG_MAX;
	lock->pid       = pid;
	lock->async     = 0;

	/*
	 * This one doesn't allocate memory, but simply copies the
	 * pointer. This is for the benefit of functions such as 
	 * lock_testlock.
	 */
	netobj_init(&(lock->oh), oh);
}

/*
 * Initialize a lock given a NLM lock request. Full-featured version.
 */
void
lk_init (lockd_lock *lock, lockd_host *host, lockd_file *file,
		u_long offset, u_long len, netobj *oh, int pid, int exclusive)
{
	lk_quick_init (lock, host, file, offset, len, oh, pid, exclusive);
	netobj_copy(&(lock->oh), oh);
}

/*
 * Find a matching lock request in a list of locks.
 */
lockd_lock *
lk_find(lockd_lock *head, nlm_lock *lkreq, int exclusive)
{
	for (; head != NULL; head = head->file.nxt) {
		if (!lk_compare(head, lkreq, exclusive))
			return head;
	}
	return NULL;
}

/*
 * Compare a lockd_lock to data from a client request.
 */
int
lk_compare (lockd_lock *lock, nlm_lock *lkreq, int exclusive)
{
	if (netobj_cmp(&(lock->oh), &(lkreq->oh)) ||
	    lock->pid != lkreq->svid ||
	    lock->l_start != lkreq->l_offset ||
	    lock->l_length != lkreq->l_len ||
	    (exclusive? !lock->exclusive : lock->exclusive))
		return 1;

	return 0;	/* that's the one */
}

/*
 * Insert lock into file list. If previous is NULL, add it at
 * the list head.
 */
void
lk_insert_file(lockd_lock **head, lockd_lock *prev, lockd_lock *lock)
{
	lockd_lock	*tmp;

	if (prev != NULL) {
		lock->file.nxt = prev->file.nxt;
		prev->file.nxt = lock;
	} else {
		lock->file.nxt = *head;
		*head = lock;
	}
	lock->file.prv = prev;

	if ((tmp = lock->file.nxt) != NULL) {
		tmp->file.prv = lock;
	}
	lock->file.linked = 1;
}

void
lk_insert_host(lockd_lock **head, lockd_lock *prev, lockd_lock *lock)
{
	lockd_lock	*tmp;

	if (prev != NULL) {
		lock->host.nxt = prev->host.nxt;
		prev->host.nxt = lock;
	} else {
		lock->host.nxt = *head;
		*head = lock;
	}
	lock->host.prv = prev;

	if ((tmp = lock->host.nxt) != NULL) {
		tmp->host.prv = lock;
	}
	lock->host.linked = 1;
}

/*
 * Append lock to a list of locks. This is not very efficient, but
 * I don't want to bother with a tail pointer for now.
 */
void 
lk_append_file(lockd_lock **head, lockd_lock *lock)
{
	lockd_lock	**prevp, *prev = NULL;

	for (prevp = head; *prevp != NULL; prevp = &((*prevp)->file.nxt)) 
		prev = *prevp;
	lk_insert_file(head, prev, lock);
}

/*
 * Free a lock. Detect automatically which list it is on.
 */
void
lk_free (lockd_lock *lock)
{
	lockd_lock	*prev, *next;
	lockd_file	*file;
	lockd_host	*host;

	if (lock->file.linked) {
		file = lock->file.back;
		prev = lock->file.prv;
		next = lock->file.nxt;
		if (prev != NULL) {
			prev->file.nxt = next;
		} else {
			if (file->granted == lock) {
				file->granted = next;
			} else if (file->blocked == lock) {
				file->blocked = next;
			} else if (file->acknowledged == lock) {
				file->acknowledged = next;
			} else {
				log(L_ERROR, "lk_free: orphan lock detected");
			}
		}
		if (next != NULL) {
			next->file.prv = prev;
		}
	}

	if (lock->host.linked) {
		host = lock->host.back;
		prev = lock->host.prv;
		next = lock->host.nxt;
		if (prev != NULL) {
			prev->host.nxt = next;
		} else {
			if (host->r_granted == lock) {
				host->r_granted = next;
			} else if (host->r_blocked == lock) {
				host->r_blocked = next;
			} else if (host->l_granted == lock) {
				host->l_granted = next;
			} else {
				log(L_ERROR, "lk_free: orphan lock detected");
			}
		}
		if (next != NULL) {
			next->host.prv = prev;
		}
	}

	netobj_free(&(lock->oh));
	xfree (lock);
}

/*
 * Remove a lock from a list.
 */
void
lk_unlink_host(lockd_lock **head, lockd_lock *lock)
{
	lockd_lock *	tmp;

	if ((tmp = lock->host.prv) != NULL) {
		tmp->host.nxt = lock->host.nxt;
	} else {
		*head = lock->host.nxt;
	}
	if ((tmp = lock->host.nxt) != NULL) {
		tmp->host.prv = lock->host.prv;
	}
	lock->host.nxt = lock->host.prv = NULL;
	lock->host.linked = 0;
}

void
lk_unlink_file(lockd_lock **head, lockd_lock *lock)
{
	lockd_lock *	tmp;

	if ((tmp = lock->file.prv) != NULL) {
		tmp->file.nxt = lock->file.nxt;
	} else {
		*head = lock->file.nxt;
	}
	if ((tmp = lock->file.nxt) != NULL) {
		tmp->file.prv = lock->file.prv;
	}
	lock->file.nxt = lock->file.prv = NULL;
	lock->file.linked = 0;
}

/*
 * Check if a new lock would clash with an existing one.
 * We first walk the list of ackowledged locks (NLM_GRANTED sent to
 * client but no reply received yet), and then the list of granted
 * locks.
 */
lockd_lock *
lk_check(lockd_lock *new, lockd_lock ***prevp)
{
	lockd_lock	*lk, **lkp;
	lockd_file	*file;
	int		pass = 0;

	if ((file = new->file.back) == NULL)
		return NULL;

	lkp = &(file->acknowledged);

again:
	pass++;

	if (prevp != NULL)
		*prevp = NULL;

	for (; *lkp != NULL; lkp = &((*lkp)->file.nxt)) {
		lk = *lkp;

		if (lk->l_end <= new->l_start)
			continue;

		if (new->l_end <= lk->l_start)
			break;

		/*
		 * Okay, we've got an overlap here. Reject any conflicting
		 * locks out of hand.
		 */
		if (lk->host.back != new->host.back || lk->pid != new->pid) {
		    	if (lk->mode & new->access || lk->access & new->mode)
				return lk;
		/* } else if (prevp != NULL && *prevp != NULL) { */
		} else if (prevp != NULL && *prevp == NULL) {
			*prevp = lkp;
		}
	}

	if (pass == 1) {
		lkp = &(file->granted);
		goto again;
	}

	return NULL;
}

/*
 * Lock or unlock a region.
 * Large parts of the code were stolen from the Linux kernel file locking
 * implementation by Kai Petzke.
 *
 * This code is terribly convoluted for several reasons:
 *
 *  -	This routine is used by the `server personality' of lockd. This
 *	means we're dealing with locks from different clients and processes.
 *	Although all locks for a single client process are made disjoint,
 *	this is not true for locks held by different processes.
 *
 *  -	We're dealing with files that may be locked or unlocked behind our
 *	backs by a local process.
 *
 *  -	This routine is used by the `client personality' of lockd as well
 *	in order to record the state of locks we hold for a given file on
 *	on a remote host.
 */
int
lk_lock(int action, lockd_lock *new)
{
	lockd_lock	*lk;
	lockd_lock	*left = NULL;
	lockd_lock	*right = NULL;
	lockd_lock	**prevp = NULL, *previous, *insafter;
	lockd_lock	**hostlist;
	lockd_file	*file;
	lockd_host	*clnt;
	int		added = 0;	/* 0 if we have to allocate a lk */
	int		local = 0;	/* 1 if file is on this host */

	file = new->file.back;
	clnt = new->host.back;

	if ((local = (file->fd >= 0)) != 0) {
		hostlist = &(clnt->r_granted);
	} else {
		hostlist = &(clnt->l_granted);
	}

	/*
	 * First, check for any conflicting locks we maintain ourselves.
	 */
	if (action != AC_UNLOCK && local && lk_check(new, &prevp) != NULL)
		return LCK_DENIED;
	if (prevp == NULL)
		prevp = &(file->granted);

	/*
	 * First, try to create the lock for real. We have to do it here
	 * to make sure we don't clash with a lock held by some other
	 * process on this machine.
	 *
	 * This is somewhat bothersome if our internal lock creation should
	 * fail later, because we have to set up all locks again. But I
	 * can't see how to do this more cleanly.
	 */
	if (action != AC_UNLOCK && local) {
		struct flock	fl;

		fl.l_type   = (new->mode == LCKMODE_SHARED)? F_RDLCK : F_WRLCK;
		fl.l_whence = SEEK_SET;
		fl.l_start  = new->l_start;
		fl.l_len    = new->l_length;
		if (fcntl(file->fd, F_SETLK, &fl) < 0)
			return LCK_DENIED;
	}

	if (action == AC_PRELOCK)
		return LCK_GRANTED;

	if (action == AC_UNLOCK)
		added = 1;

	/*
	 * If we get here, there are no conflicting locks. Go ahead and
	 * insert it.
	 */
	previous = insafter = NULL;
	while ((lk = *prevp) != NULL) {
		if (lk->l_end <= new->l_start)
			goto next_lock;

		if (new->l_end <= lk->l_start)
			break;

		/*
		 * Okay, we've got an overlap here. Only locks held by the
		 * same process need apply.
		 */
		if (lk->host.back != new->host.back || lk->pid != new->pid)
			goto next_lock;

		if (action != AC_UNLOCK && new->mode == lk->mode) {
			/*
			 * If we come here, the new and old lock are of the
			 * same type and adjacent or overlapping. Make one
			 * lock yielding from the lower start address of both
			 * locks to the higher end address.
			 */
			if (lk->l_start > new->l_start)
				lk->l_start = new->l_start;
			else
				new->l_start = lk->l_start;
			if (lk->l_end < new->l_end)
				lk->l_end = new->l_end;
			else
				new->l_end = lk->l_end;
			if (lk->l_length == 0)
				new->l_length = lk->l_length;
			else
				lk->l_length = lk->l_end - lk->l_start;
			if (new->l_length == 0)
				lk->l_length = new->l_length;
			else
				new->l_length = new->l_end - new->l_start;

			/*
			 * If we have already added the new lock, discard
			 * the current one - it's become simply a duplicate.
			 */
			if (added) {
				lk_free(lk);
				continue;
			}
			new = lk;
			added = 1;
			goto next_lock;
		}

		/*
		 * These locks have different type, and overlap. This means
		 * part of the old lock will be replaced with the new type
		 * of lock.
		 */

		/*
		 * If the new lock starts after or ends before the current
		 * one, remember that lock as `left' or `right', resp.
		 */
		if (lk->l_start < new->l_start)
			left = lk;

		if (new->l_end < lk->l_end) {
			right = lk;
			break;
		}

		if (new->l_start <= lk->l_start) {
			/*
			 * The new lock completely replaces an old one (This may
			 * happen several times).
			 */
			if (action == AC_UNLOCK) {
				file->retry_blocked = 1;
			}
			if (added) {
				lk_free(lk);
				continue;
			}

			/*
			 * Replace the old lock with the new one.
			 */
			lk->l_start   = new->l_start;
			lk->l_end     = new->l_end;
			lk->l_length  = new->l_length;
			lk->mode      = new->mode;
			lk->access    = new->access;
			lk->exclusive = new->exclusive;
			new = lk;
			added = 1;
		}
		/*
		 * Go on to next lock.
		 */
next_lock:
		if (lk->l_start <= new->l_start)
			insafter = lk;
		previous = lk;
		prevp = &(lk->file.nxt);
	}

	if (right) {
		if (left == right) {
			/*
			 * The new lock breaks the old one in two pieces, so we
			 * have to allocate one more lock.
			 */
			if ((left = lk_dup (left)) == NULL)
				return LCK_DENIED_NOLOCKS;
			lk_insert_file(&(file->granted), previous, left);
			lk_insert_host(hostlist, NULL, left);
		}
		right->l_start = new->l_end;
		if (right->l_length != 0)
			right->l_length = right->l_end - right->l_start;
	}
	if (left) {
		left->l_length = new->l_start - left->l_start;
		left->l_end = new->l_start;
	}
	if (! added) {
		if (action == AC_UNLOCK)
			return LCK_GRANTED;
		if ((new = lk_dup(new)) == NULL)
			return LCK_DENIED_NOLOCKS;
		if (left) {
			lk_insert_file(&(file->granted), left, new);
		} else {
			lk_insert_file(&(file->granted), insafter, new);
		}
		lk_insert_host(hostlist, NULL, new);
	}

	/*
	 * Alright, now try to unlock all those regions not covered.
	 */
	if (local) {
		if (action == AC_UNLOCK)
			lk_update(file, new->l_start, new->l_end, AC_UNLOCK);
		else /* for debugging: */
			lk_update(file, new->l_start, new->l_end, action);
	}

	return LCK_GRANTED;
}

/*
 * Update all locks that lie within a given region by either unlocking
 * newly opened gaps in the file or locking newly added/modified regions
 * within the file. It relies on locks being sorted in order of ascending
 * l_start addresses.
 *
 * What makes the code particularly nasty is that we have to walk the
 * granted and acknowledged lists in parallel.
 */
void
lk_update(lockd_file *file, unsigned long start, unsigned long end, int action)
{
	unsigned long	lkstart = 0, lklen = 0, lkend = 0;
	int		lkexcl;
	lockd_lock	*lk, *glk, *alk;

	lkstart = lkend = start;
	lklen   = 0;
	lkexcl  = 0;

	glk = file->granted;
	alk = file->acknowledged;
	while (glk != NULL || alk != NULL) {

		if (glk == NULL || (alk && alk->l_start < glk->l_start)) {
			lk = alk; alk = alk->file.nxt;
		} else {
			lk = glk; glk = glk->file.nxt;
		}

		if (action == AC_UNLOCK && lk->l_length == 0)
			break;

		if (lk->l_end <= start)
			continue;

		if (end != 0 && end <= lk->l_start)
			break;

		if (lk->l_start < lkstart) {
			log(L_ERROR, "internal error: misordered lock list!");
			continue;
		}

		if (lk->l_start < lkend) {
			/* We have a lock overlap; Extend the current lock
			 * region.
			 */
			if (lk->exclusive != lkexcl) {
				log(L_ERROR,
					"internal error: ex/sh lock overlap!");
				lkexcl = 1;
			}
			if (lk->l_end > lkend) {
				lkend = lk->l_end;
				lklen = lk->l_length? lkend - lkstart : 0;
			}
			continue;
		}

		if (action == AC_UNLOCK) {
			if (lk->l_start != lkend)
				unlock_region(file, lkend, lk->l_start - lkend);
		} else {
			if (lkstart != lkend)
				lock_region(file, lkstart, lklen, lkexcl);
		}

		lkstart = lk->l_start;
		lkend   = lk->l_end;
		lklen   = lk->l_length;
		lkexcl  = lk->exclusive;
	}
	if (action == AC_UNLOCK) {
		if (lk == NULL) {
			unlock_region(file, lkend, 0);
		} else {
			unlock_region(file, lkend, lk->l_start - lkend);
		}
	} else if (lkstart != lkend) {
		lock_region(file, lkstart, lklen, lkexcl);
	}
}

void
unlock_region(lockd_file *file, unsigned long start, unsigned long len)
{
	struct flock	fl;

	fl.l_whence = SEEK_SET;
	fl.l_type   = F_UNLCK;
	fl.l_start  = start;
	fl.l_len    = len;
	fcntl(file_fd(file), F_SETLK, &fl);
	log(L_DEBUG, "unlocked region [%d:%d] of %s",
			start, len, 
			file_pathname(file));
}

void
lock_region(lockd_file *file, unsigned long start, unsigned long len, int excl)
{
	struct flock	fl;

	fl.l_whence = SEEK_SET;
	fl.l_type   = excl?  F_WRLCK : F_RDLCK;
	fl.l_start  = start;
	fl.l_len    = len;
	if (fcntl(file_fd(file), F_SETLK, &fl) < 0) {
		log(L_ERROR, "failed to %slock region [%d:%d] of %s",
				excl? "ex" : "sh",
				start, len, 
				file_pathname(file));
	} else {
		log(L_DEBUG, "%slocked region [%d:%d] of %s",
				excl? "ex" : "sh",
				start, len, 
				file_pathname(file));
	}
}

/*
 * Debug functionality: dump locks
 */
void
lk_dumpfile(lockd_file *file)
{
	lockd_lock	*lk;

	printf("\nfile %s\n ", file->name? file->name : "<remote>");
	if (file->granted != NULL)
		printf("    granted:\n ");
	for (lk = file->granted; lk != NULL; lk = lk->file.nxt) {
		lk_dumplock(lk);
	}
	if (file->acknowledged != NULL)
		printf("    acknowledged:\n ");
	for (lk = file->acknowledged; lk != NULL; lk = lk->file.nxt) {
		lk_dumplock(lk);
	}
	if (file->blocked != NULL)
		printf("    blocked:\n ");
	for (lk = file->blocked; lk != NULL; lk = lk->file.nxt) {
		lk_dumplock(lk);
	}
}

void
lk_dumplock(lockd_lock *lk)
{
	printf("\tstart/len/end:\t%ld/%ld/%ld\n"
	       "\tmode/access:\t%d/%d\n"
	       "\tpid:\t\t%d\n",
	       lk->l_start, lk->l_length, lk->l_end,
	       lk->mode, lk->access,
	       lk->pid);
}

