/*
 * lock_s.c	Server side lock handling.
 *
 * Copyright (C) 1995, Olaf Kirch <okir@monad.swb.de>
 */

#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include "nlm.h"

/*
 * Check if a lock request would clash with an existing lock.
 */
lockd_lock *
lock_testlock(lockd_host *clnt, lockd_file *file, nlm_lock *lkreq, int excl)
{
	lockd_lock	lock;

	lk_quick_init(&lock, clnt, file, lkreq->l_offset, lkreq->l_len,
				&(lkreq->oh), lkreq->svid, excl);
	return lk_check(&lock, NULL);
}

int
lock_createlock(lockd_host *clnt, lockd_file *file, nlm_lock *lkreq, int excl)
{
	lockd_lock	lock;
	int		status;

	lk_quick_init(&lock, clnt, file, lkreq->l_offset, lkreq->l_len,
				&(lkreq->oh), lkreq->svid, excl);
	status = lk_lock(AC_LOCK, &lock);

	/*
	 * If the lock operation failed, we must unlock all erroneously
	 * claimed regions.
	 */
	if (status == LCK_DENIED)
		lk_update(lock.file.back, lock.l_start, lock.l_end, AC_UNLOCK);

	return status;
}

int
lock_unlock(lockd_host *clnt, lockd_file *file, nlm_lock *lkreq)
{
	lockd_lock	lock;

	lk_quick_init(&lock, clnt, file, lkreq->l_offset, lkreq->l_len,
				&(lkreq->oh), lkreq->svid, 0);
	file->retry_blocked = 0;
	return lk_lock(AC_UNLOCK, &lock);
}

int
lock_createshare(lockd_host *clnt, lockd_file *file, nlm_share *sh)
{
	lockd_lock	lock;
	int		status;

	lk_quick_init(&lock, clnt, file, 0, 0, &sh->oh, 1, (sh->mode & fsm_DW));
	lock.mode = sh->mode;
	lock.access = sh->access;
	status = lk_lock(AC_LOCK, &lock);

	/*
	 * If the lock operation failed, we must unlock all erroneously
	 * claimed regions.
	 */
	if (status == LCK_DENIED)
		lk_update(file, 0, 0, AC_UNLOCK);

	return status;
}

int
lock_unshare(lockd_host *clnt, lockd_file *file, nlm_share *sh)
{
	lockd_lock	lock;

	lk_quick_init(&lock, clnt, file, 0, 0, &sh->oh, 1, (sh->mode & fsm_DW));
	lock.mode = sh->mode;
	lock.access = sh->access;
	file->retry_blocked = 0;
	return lk_lock(AC_UNLOCK, &lock);
}

void
lock_unlock_all(lockd_host *clnt)
{
	lock_freeall(&(clnt->r_granted));
	lock_freeall(&(clnt->r_blocked));
}

/*
 * Delete all locks on a host's queue.
 */
void
lock_freeall(lockd_lock **head)
{
	lockd_lock	*lock, *next;

	while ((lock = *head) != NULL) {
		next = lock->host.nxt;
		lock->file.back->update_locks = 1;
		lock->file.back->retry_blocked = 1;
		lk_free (lock);
	}
}

/*
 * Add a client request to a host's blocking list. Duplicate requests
 * will be ignored.
 */
void
lock_block_clntreq(lockd_host *clnt, lockd_file *file, nlm_lock *req, 
					int excl, int async)
{
	lockd_lock	*lk;

	if (lk_find(clnt->r_blocked, req, excl))
		return;

	lk = lk_alloc(NULL, NULL); /* lk_init zeroes lock struct anyway */
	lk_init(lk, clnt, file, req->l_offset, req->l_len, 
					&(req->oh), req->svid, excl);
	lk->async = async;

	lk_insert_host(&(clnt->r_blocked), NULL, lk);
	lk_append_file(&(file->blocked), lk);
}

/*
 * Retry a blocked client request. This code is somewhat hairy because
 * we have to take care of the following scenario: Assume we grant the
 * lock request right away, and insert it into the lock list. Then we
 * tell the client the lock has been granted, but either the RPC call
 * fails or the client is currently unable to process it (the X/Open
 * Spec explicitly mentions grace period here). We then have to re-queue
 * the lock request in the queue of blocked requests and restore the
 * original lock list.
 *
 * We do this by first locking the file with the kernel, then placing
 * the GRANT request. If the call returns OK, we also insert the request
 * in our own lock list. If it is reported as in-progress (as happens
 * with the asynchronous NLM_GRANTED_MSG call), we put it on the list
 * of acked-but-not-granted requests and return. If the GRANT operation
 * fails, we move the lock request to the end of blocked requests and
 * restore the original set of locks with the kernel.
 */
int
lock_grant_clntreq(lockd_lock *lk)
{
	lockd_file	*file = lk->file.back;
	lockd_host	*clnt = lk->host.back;

	if (lk_lock(AC_PRELOCK, lk) != LCK_GRANTED) {
		return LCK_DENIED;
	}

	lk_unlink_file(&(file->blocked), lk);
	lk_unlink_host(&(clnt->r_blocked), lk);

	switch (nlm_grant(lk)) {
	case RQ_QUEUED:
		log(L_DEBUG, "queued grant notification");
		lk_insert_file(&(file->acknowledged), NULL, lk);
		lk_insert_host(&(clnt->r_granted), NULL, lk);
		return LCK_GRANTED;
	case RQ_GRANTED:
		if (lk_lock(AC_LOCK, lk) == LCK_GRANTED) {
			lk_free(lk);
			return LCK_GRANTED;
		}
		log(L_ERROR, "internal error in lock_grant_clntreq");
	}
	/* repair file locks */
	lk_update(file, lk->l_start, lk->l_end, AC_UNLOCK);

	log(L_DEBUG, "re-blocking lock request");
	lk_append_file(&(file->blocked), lk);
	lk_insert_host(&(clnt->r_blocked), NULL, lk);
	return LCK_DENIED;
}

/*
 * This is the callback invoked when the RPC reply from an asynchronous
 * GRANT callback to the client is received. If we arrive here, the
 * lock should be on the file's acknowledged and the host's granted
 * list.
 */
void
lock_grant_status(int status, lockd_lock * lock)
{
	lockd_host *	host = lock->host.back;
	lockd_file *	file = lock->file.back;

	lk_unlink_file(&(file->acknowledged), lock);
	lk_unlink_host(&(host->r_granted), lock);

	if (status == RQ_GRANTED) {
		log(L_DEBUG, "committing granted lock request");
		if (lk_lock(AC_LOCK, lock) == LCK_GRANTED) {
			lk_free(lock);
			return;
		}
		log(L_ERROR, "internal error in lock_grant_clnt");
	}
	/* repair file locks */
	lk_update(file, lock->l_start, lock->l_end, AC_UNLOCK);

	log(L_DEBUG, "re-blocking lock request (grant failed)");
	lk_append_file(&(file->blocked), lock);
	lk_insert_host(&(host->r_blocked), NULL, lock);
}

/*
 * Cancel a client's outstanding lock request.
 * FIXME: What happens if cancel request arrives after we sent out the
 * grant? For now, we just remove it and update the locks, but I'm not
 * sure if that's sufficient.
 */
void
lock_cancel_clntreq(lockd_host *clnt, lockd_file *file, nlm_lock *req, int excl)
{
	lockd_lock	*lk;

	if ((lk = lk_find(clnt->r_blocked, req, excl)) != NULL) {
		lk_free(lk);
	} else if ((lk = lk_find(clnt->r_granted, req, excl)) != NULL) {
		lk_update(file, lk->l_start, lk->l_end, AC_UNLOCK);
		lk_free(lk);
	}
}

/*
 * Dump all locks.
 */
void
lock_dump(void)
{
	file_foreach(lk_dumpfile);
}
