/*
 * nlm.c	These routines implement the server side of the NLM
 *		protocol, except for lock granting which is handled in
 *		nlm_grant.c.
 *
 * Copyright (C) 1995, Olaf Kirch <okir@monad.swb.de>
 */

#include <rpc/rpc.h>
#include <arpa/inet.h>
#include "nlm.h"
#include "sm_inter.h"

/*
 * Global variables.
 */
static int	grace = 0;	/* 1 during grace period */

/*
 * These functions are the generic part for both synchronous and asynchronous
 * LOCK, UNLOCK etc.
 */
static bool_t	nlm_test_impl(struct svc_req *, struct nlm_testargs *,
							nlm_testres *);
static bool_t	nlm_lock_impl(struct svc_req *, struct nlm_lockargs *,
							nlm_res *, int);
static bool_t	nlm_cancel_impl(struct svc_req *, struct nlm_cancargs *,
							nlm_res *);
static bool_t	nlm_unlock_impl(struct svc_req *, struct nlm_unlockargs *,
							nlm_res *);
static bool_t	nlm_grant_impl(struct svc_req *, struct nlm_testargs *,
							nlm_res *);
static int	nlm_result_impl(struct svc_req *rqstp, void * argp);
static int	nlm_rqstatus(int nlmstat);
static void	nlm_prlock(struct svc_req *, nlm_lock *lk, int, int);
static void	nlm_prshare(struct svc_req *, nlm_share *sh);


/*
 * NULL procedure. Nop.
 */
bool_t
nlm_null_svc (struct svc_req *rqstp, void *argp, void *resp)
{
	/* nothing happens */
	return 1;
}

/*
 * Test for presence of conflicting lock
 */
bool_t
nlm_test_svc (struct svc_req * rqstp, nlm_testargs * argp, nlm_testres * resp)
{
	int	ok;

	if ((ok = nlm_test_impl(rqstp, argp, resp))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * Create a lock
 */
bool_t
nlm_lock_svc (struct svc_req * rqstp, nlm_lockargs * argp, nlm_res * resp)
{
	int	ok;

	if ((ok = nlm_lock_impl(rqstp, argp, resp, 1))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * Cancel a blocked lock request.
 */
bool_t
nlm_cancel_svc (struct svc_req * rqstp, nlm_cancargs * argp, nlm_res * resp)
{
	int	ok;

	if ((ok = nlm_cancel_impl(rqstp, argp, resp))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * Free a lock.
 */
bool_t
nlm_unlock_svc (struct svc_req * rqstp, nlm_unlockargs * argp, nlm_res * resp)
{
	int	ok;

	if ((ok = nlm_unlock_impl(rqstp, argp, resp))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * The server grants a previously blocked lock.
 */
bool_t
nlm_granted_svc (struct svc_req *rqstp, nlm_testargs * argp, nlm_res * resp)
{
	int		ok;

	if ((ok = nlm_grant_impl(rqstp, argp, resp))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * Test for conflicting lock, asynchronous version.
 */
bool_t
nlm_test_msg_svc (struct svc_req * rqstp, nlm_testargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_testres	res;

	if ((host = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if (!nlm_test_impl(rqstp, argp, &res)) {
		return 0;
	}
	netobj_init(&(res.cookie), &(argp->cookie));

	rpcmsg_send(host->addr, nlm_cltpg, NLM_VERS, NLM_TEST_RES,
				(xdrproc_t) xdr_nlm_testres, &res);

	return 1;
}

/*
 * Create a lock, asynchronous version.
 */
bool_t
nlm_lock_msg_svc (struct svc_req * rqstp, nlm_lockargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_res		res;

	if ((host = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if (!nlm_lock_impl(rqstp, argp, &res, 1)) {
		return 0;
	}
	netobj_init(&(res.cookie), &(argp->cookie));

	rpcmsg_send(host->addr, nlm_cltpg, NLM_VERS, NLM_LOCK_RES,
				(xdrproc_t) xdr_nlm_res, &res);

	return 1;
}

/*
 * Cancel a lock, asynchronous version.
 */
bool_t
nlm_cancel_msg_svc (struct svc_req * rqstp, nlm_cancargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_res		res;

	if ((host = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if (!nlm_cancel_impl(rqstp, argp, &res)) {
		return 0;
	}
	netobj_init(&(res.cookie), &(argp->cookie));

	rpcmsg_send(host->addr, nlm_cltpg, NLM_VERS, NLM_CANCEL_RES,
				(xdrproc_t) xdr_nlm_res, &res);

	return 1;
}

/*
 * Unlock a lock, asynchronous version.
 */
bool_t
nlm_unlock_msg_svc (struct svc_req * rqstp, nlm_unlockargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_res		res;

	if ((host = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if (!nlm_unlock_impl(rqstp, argp, &res)) {
		return 0;
	}
	netobj_init(&(res.cookie), &(argp->cookie));

	rpcmsg_send(host->addr, nlm_cltpg, NLM_VERS, NLM_UNLOCK_RES,
				(xdrproc_t) xdr_nlm_res, &res);

	return 1;
}

/*
 * Server grants a previously blocked lock (async version).
 */
bool_t
nlm_granted_msg_svc(struct svc_req * rqstp, nlm_testargs * argp, void * nresp)
{
	lockd_host *	host;
	nlm_res		res;
	int		status;

	if ((host = host_lookup(rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	if (!nlm_grant_impl(rqstp, argp, &res)) {
		return 0;
	}
	netobj_init(&res.cookie, &argp->cookie);

	status = rpcmsg_send(host->addr, nlm_svcpg, NLM_VERS, NLM_GRANTED_RES,
				(xdrproc_t) xdr_nlm_res, &res);

	return 1;
}

/*
 * The server returns the result of a TEST_MSG call.
 */
bool_t
nlm_test_res_svc (struct svc_req * rqstp, nlm_testres * argp, void * resp)
{
	return nlm_result_impl(rqstp, argp);
}

/*
 * Server returns the result of a LOCK_MSG call.
 * Before returning the result to the original caller, we have to add
 * the lock to our own list of lock requests so that we're able to do
 * lock recovery.
 */
bool_t
nlm_lock_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_result_impl(rqstp, argp);
}

/*
 * Server returns the result of a CANCEL_MSG call.
 */
bool_t
nlm_cancel_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_result_impl(rqstp, argp);
}

/*
 * Server returns the result of an UNLOCK_MSG call.
 */
bool_t
nlm_unlock_res_svc (struct svc_req * rqstp, nlm_res * argp, void * resp)
{
	return nlm_result_impl(rqstp, argp);
}

/*
 * Result of a GRANT call
 */
bool_t
nlm_granted_res_svc(struct svc_req *rqstp, nlm_res *argp, void *resp)
{
	lockd_lock	*lock;
	int		cookie, status;

	if (!netobj_get(&(argp->cookie), &cookie, sizeof(cookie))) {
		svcerr_decode(rqstp->rq_xprt);
		return 0;
	}
	if ((lock = nlm_dequeue_grant(cookie)) == NULL) {
		log(L_WARNING, "grant_res for unknown lock");
		svcerr_systemerr(rqstp->rq_xprt);
		return 0;
	}

	status = (argp->stat.stat == LCK_GRANTED)? RQ_GRANTED : RQ_FAILED;
	lock_grant_status(status, lock);

	return 1;
}

bool_t
nlm_share_svc(struct svc_req *rqstp, nlm_shareargs *argp, nlm_shareres *resp)
{
	lockd_host *clnt;
	lockd_file *file;

	/* NLM spec says cookie & sequence should be ignored */
	netobj_init(&resp->cookie, &argp->cookie);
	resp->sequence = 0;

	if (grace && !argp->reclaim) {
		resp->stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((clnt = host_lookup(rqstp, argp->share.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if ((file = file_lookup(clnt, &(argp->share.fh), 1)) == NULL) {
		if (errno == EMFILE)
			resp->stat = LCK_DENIED_NOLOCKS;
		else
			resp->stat = LCK_DENIED;
		return 1;
	}

	log(L_DEBUG, "share %s", file->name);
	nlm_prshare(rqstp, &argp->share);

	/* Try to create the lock */
	resp->stat = lock_createshare(clnt, file, &(argp->share));

	/* A non-exclusive lock request might remove an exclusive lock,
	 * so retry.
	 */
	if (resp->stat == LCK_GRANTED && !(argp->share.mode & fsm_DW)) {
		file_retry_blocked (file);
	}
	file_release (clnt, file);

	return 1;
}

bool_t
nlm_unshare_svc(struct svc_req *rqstp, nlm_shareargs *argp, nlm_shareres *resp)
{
	lockd_host *clnt;
	lockd_file *file;

	/* NLM spec says cookie & sequence should be ignored */
	netobj_init(&resp->cookie, &argp->cookie);
	resp->sequence = 0;

	if (grace) {
		resp->stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((clnt = host_lookup (rqstp, argp->share.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if ((file = file_lookup (clnt, &(argp->share.fh), 1)) == NULL) {
		if (errno == EMFILE) {
			resp->stat = LCK_DENIED_NOLOCKS;
			return 1;
		}
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	log(L_DEBUG, "unshare %s", file->name);
	nlm_prshare(rqstp, &argp->share);

	/* Try to remove the lock */
	resp->stat = lock_unshare(clnt, file, &argp->share);

	if (file->retry_blocked) {
		file_retry_blocked(file);
	}

	/* If there are no more locks on this file, delete it */
	file_release (clnt, file);

	return 1;
}


/*
 * Create an unmonitored lock
 */
bool_t
nlm_nm_lock_svc (struct svc_req * rqstp, nlm_lockargs * argp, nlm_res * resp)
{
	int	ok;

	if ((ok = nlm_lock_impl(rqstp, argp, resp, 0))) {
		netobj_init (&(resp->cookie), &(argp->cookie));
	}

	return ok;
}

/*
 * A client has just rebooted and gives up all its locks.
 */
bool_t
nlm_free_all_svc (struct svc_req * rqstp, nlm_notify * argp, void * resp)
{
	lockd_host *clnt;

	if ((clnt = host_lookup (rqstp, argp->name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}
	lock_unlock_all(clnt);
	nlm_unmonitor(clnt);

	return 1;
}

/*
 * This RPC is not specified by NLM, but we need it as a callback for
 * NSM.
 */
bool_t
nlm_reclaim_svc(struct svc_req *rqstp, nsm_status *argp, void *resp)
{
	lockd_host *	host;

	if ((host = host_lookup(rqstp, argp->mon_name, 0)) == NULL)
		return 1;
	/* Guard against NSM retransmissions */
	if (host->state != argp->state) {
		nlm_reclaim_all(host, argp->state);
	}
	return 1;
}

/*
 * Here comes the implementation of all those beautiful NLM server side
 * functions. These functions are invoked by both the synchronous and
 * asynchronous versions.
 */
static bool_t
nlm_test_impl(struct svc_req * rqstp, nlm_testargs * argp, nlm_testres * resp)
{
	lockd_lock *lock;
	lockd_file *file;
	lockd_host *host;

	/* We won't grant new locks during grace period */
	if (grace) {
		resp->stat.stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((host = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		netobj_init (&(resp->cookie), &(argp->cookie));
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	/* Get the file handle */
	if ((file = file_lookup (host, &(argp->alock.fh), 1)) == NULL) {
		if (errno == EMFILE) {
			resp->stat.stat = LCK_DENIED_NOLOCKS;
			return 1;
		}
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	/* Check for conflicting locks */
	lock = lock_testlock (host, file, &(argp->alock), argp->exclusive);
	if (lock != NULL) {
		nlm_holder *holder;

		holder = &(resp->stat.nlm_testrply_u.holder);
		netobj_init (&(holder->oh), &(lock->oh));
		holder->exclusive = !(lock->mode == LCKMODE_SHARED);
		holder->svid = lock->pid;
		holder->l_offset = lock->l_start;
		holder->l_len = lock->l_length;
		resp->stat.stat = LCK_DENIED;
	} else {
		resp->stat.stat = LCK_GRANTED;
	}
	file_release (host, file);

	return 1;
}

/*
 * Create a lock.
 */
static bool_t
nlm_lock_impl(struct svc_req *rqstp, nlm_lockargs *argp, nlm_res *resp,
							int monitor)
{
	lockd_host *clnt;
	lockd_file *file;

	if (grace && !argp->reclaim) {
		resp->stat.stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((clnt = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	/* Check if the state matches. We can either remove all locks if
	 * the passed state is different from current, or fail the request.
	 */
	if (clnt->monitored && clnt->state != argp->state) {
		if (clnt->state == 0) {
			clnt->state = argp->state;
		} else {
			log(L_WARNING,
				"state %d in lock request from %s (have %d)\n",
				argp->state, clnt->name, clnt->state);
			if (clnt->state < argp->state) {
				lock_unlock_all(clnt);
			} else {
				resp->stat.stat = LCK_DENIED;
				return 1;
			}
		}
	}

	if ((file = file_lookup (clnt, &(argp->alock.fh), 1)) == NULL) {
		if (errno == EMFILE)
			resp->stat.stat = LCK_DENIED_NOLOCKS;
		else
			resp->stat.stat = LCK_DENIED;
		return 1;
	}

	log(L_DEBUG, "lock %s%s", file->name, monitor? "" : " (unmon)");
	nlm_prlock(rqstp, &argp->alock, argp->block, argp->exclusive);

	/* Try to create the lock */
	resp->stat.stat = lock_createlock (clnt, file, &(argp->alock),
				argp->exclusive);

	/* If we failed but the client requested blocking, we queue the
	 * request, or fail it for reclaimed locks.
	 */
	if (resp->stat.stat == LCK_DENIED && argp->block) {
		if (!argp->reclaim) {
			lock_block_clntreq (clnt, file, &(argp->alock),
						argp->exclusive, 0);
			resp->stat.stat = LCK_BLOCKED;
		} else {
			resp->stat.stat = LCK_DENIED;
		}
	}

	/* A non-exclusive lock request might remove an exclusive lock,
	 * so retry.
	 */
	if (resp->stat.stat == LCK_GRANTED && !argp->exclusive) {
		file_retry_blocked (file);
	}
	file_release (clnt, file);

	if (monitor && resp->stat.stat == LCK_GRANTED)
		nlm_monitor (clnt, argp->state);

	return 1;
}

/*
 * A client requests that a lock request be cancelled.
 */
static bool_t
nlm_cancel_impl(struct svc_req * rqstp, nlm_cancargs * argp, nlm_res * resp)
{
	lockd_host *clnt;
	lockd_file *file;

	if (grace) {
		resp->stat.stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((clnt = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if ((file = file_lookup (clnt, &(argp->alock.fh), 1)) == NULL) {
		if (errno == EMFILE) {
			resp->stat.stat = LCK_DENIED_NOLOCKS;
			return 1;
		}
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	log(L_DEBUG, "cancel %s", file->name);
	nlm_prlock(rqstp, &argp->alock, 0, argp->exclusive);

	lock_cancel_clntreq (clnt, file, &(argp->alock), argp->exclusive);

	/* If there are no more locks on this file, delete it */
	file_release (clnt, file);

	/* If the client doesn't hold any more locks, unmonitor it */
	nlm_unmonitor (clnt);

	resp->stat.stat = LCK_GRANTED;
	return 1;
}

/*
 * Free a lock.
 */
static bool_t
nlm_unlock_impl(struct svc_req * rqstp, nlm_unlockargs * argp, nlm_res * resp)
{
	lockd_host *clnt;
	lockd_file *file;

	if (grace) {
		resp->stat.stat = LCK_DENIED_GRACE_PERIOD;
		return 1;
	}

	if ((clnt = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	if ((file = file_lookup (clnt, &(argp->alock.fh), 1)) == NULL) {
		if (errno == EMFILE) {
			resp->stat.stat = LCK_DENIED_NOLOCKS;
			return 1;
		}
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	log(L_DEBUG, "unlock %s", file->name);
	nlm_prlock(rqstp, &argp->alock, 0, 0);

	/* Try to remove the lock */
	resp->stat.stat = lock_unlock (clnt, file, &(argp->alock));

	if (file->retry_blocked) {
		file_retry_blocked (file);
	}

	/* If there are no more locks on this file, delete it */
	file_release (clnt, file);

	/* If the client doesn't hold any more locks, unmonitor it */
	nlm_unmonitor (clnt);

	return 1;
}

/* 
 * The server has granted a previously blocked request. 
 * FIXME: Should we deny GRANTs when in grace period? Don't think so...
 */
static bool_t
nlm_grant_impl(struct svc_req *rqstp, nlm_testargs * argp, nlm_res *resp)
{
        struct lklm_lock        klock, *alock = &klock;
	lockd_host *		clnt;
        lockd_req * 		rq;
        nlm_lock *		nlock = &(argp->alock);
        int			ex = argp->exclusive;

	if ((clnt = host_lookup (rqstp, argp->alock.caller_name, 1)) == NULL) {
		svcerr_systemerr (rqstp->rq_xprt);
		return 0;
	}

	log(L_DEBUG, "grant <remote>");
	nlm_prlock(rqstp, &argp->alock, 0, ex);

	if ((klock.fh.len = nlock->fh.n_len) > sizeof(klock.fh.data))
		return 0;
	memcpy(klock.fh.data, nlock->fh.n_bytes, klock.fh.len);
        strncpy(klock.server_name, clnt->name, sizeof(klock.server_name));
        klock.l_offset = nlock->l_offset;
        klock.l_len = nlock->l_len;
        klock.pid = nlock->svid;

        /* Check if there's a pending lock request that is still waiting
         * for its reply/callback. This may happen when the server's 
         * reply got lost, but the grant notification didn't.
         */
        if ((rq = req_find(clnt->l_waiting, RQ_LOCK, alock, ex, -1)) != NULL) {
        	goto foundit;
        }

        /* Find the corresponding blocked lock. If we don't find it, return 0
         * to make the caller return an error to the granting NLM server.
         * It is important for the server to know that a grant request failed,
         * otherwise the granted lock won't be owned by anyone.
         */
        if ((rq = req_find(clnt->l_blocked, RQ_LOCK, alock, ex, -1)) == NULL) {
		log(L_WARNING, "%s granted unknown lock", clnt->name);
		resp->stat.stat = LCK_DENIED;
                return 1;
        }

foundit:
        /* Found it. */
        /* req_unlink(rq); This is done in req_process_result */

	/* Update list of local locks */
	nlm_update_local(AC_LOCK, rq);

	/* Propagate the result to the client */
	req_process_result(rq, RQ_GRANTED);

	resp->stat.stat = LCK_GRANTED;
        return 1;
}

static int
nlm_result_impl(struct svc_req *rqstp, void * argp)
{
	nlm_res *	resp = (nlm_res *) argp;
	lockd_req *	rq;
	int		status, cookie;

	if (!netobj_get(&(resp->cookie), &cookie, sizeof(cookie))) {
		svcerr_decode(rqstp->rq_xprt);
		return 0;
	}

	if ((rq = req_match(cookie)) == NULL) {
		log(L_DEBUG, "received NLM result for unknown cookie %d",
						cookie);
		return 1;
	}

	status = resp->stat.stat;
	rq->result = argp;

	if (rq->type == RQ_LOCK && status == LCK_GRANTED)
		nlm_update_local(AC_LOCK, rq);
	else if (rq->type == RQ_UNLOCK && status == LCK_GRANTED)
		nlm_update_local(AC_UNLOCK, rq);

	req_process_result(rq, nlm_rqstatus(status));

	return 1;
}

/*
 * The server has accepted a lock or unlock request. Update our local
 * list of locks.
 */
void
nlm_update_local(int action, struct lockd_req *rq)
{
	lockd_lock	lock;
	lockd_file *	file;

	if (rq == NULL) return;

	if ((file = file_lookup(rq->host, &(rq->fh), 0)) == NULL) {
		log(L_ERROR, "couldn't create file struct for remote file");
		goto bonkers;
	}

	lk_quick_init(&lock, rq->host, file, rq->l_offset, rq->l_len,
				&rq->oh, rq->pid, rq->exclusive);
	if (lk_lock(action, &lock) != LCK_GRANTED) {
		log(L_ERROR, "couldn't add remote lock to local list");
		goto bonkers;
	}

	if (action == AC_LOCK)
		nlm_monitor(rq->host, -1);
	else if (action == AC_UNLOCK)
		nlm_unmonitor(rq->host);
	return;

bonkers:
	/* FIXME: should do something useful here. */
}

static int
nlm_rqstatus(int nlmstat)
{
static struct {
	int	nlmstat;
	int	rqstat;
} statusmap[] = {
	{ LCK_GRANTED,		RQ_GRANTED },
	{ LCK_DENIED,		RQ_DENIED },
	{ LCK_DENIED_NOLOCKS,	RQ_DENIED_NOLOCKS },
	{ LCK_BLOCKED,		RQ_BLOCKED },
	{ LCK_DENIED_GRACE_PERIOD, RQ_DENIED_GRACE_PERIOD },
	{ -1,			RQ_FAILED },
};
	int	i;

	for (i = 0; statusmap[i].nlmstat >= 0; i++)
		if (statusmap[i].nlmstat == nlmstat)
			break;

	return statusmap[i].rqstat;
}

static void
nlm_prlock(struct svc_req *rqstp, nlm_lock *lk, int block, int excl)
{
	struct in_addr	addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;

	log(L_DEBUG, "    %s%s[%d,%d] by pid %d@%s\n",
			excl? "excl " : "",
			block? "block " : "",
			lk->l_offset, lk->l_len, 
			lk->svid, inet_ntoa(addr));
}

static void
nlm_prshare(struct svc_req *rqstp, nlm_share *sh)
{
	struct in_addr	addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
	int		mode = sh->mode;
	int		access = sh->access;

	log(L_DEBUG, "    D%s A%s by DOS box %s\n",
			((mode == fsm_DN)? "N" : 
			 (mode == fsm_DR)? "R" : 
			 (mode == fsm_DW)? "W" : "RW"),
			((access == fsa_NONE)? "N" : 
			 (access == fsa_R)? "R" : 
			 (access == fsa_W)? "W" : "RW"),
			inet_ntoa(addr));
}
