/*
 * nlm_client.c		These functions implement the client side of the
 *			NLM server.
 *
 * Requests to be sent via NLM are wrapped up in lockd_req structs. The
 * bookkeeping for these structs is done almost entirely by klm.c and
 * request.c. The reason to separate this is that I want to be able to
 * support other locking protocols, too. (Don't know what SMB does, but it
 * should have something like this. Maybe DOS file sharing or some such).
 *
 * Copyright (C) 1995, Olaf Kirch  <okir@monad.swb.de>
 */

#include <arpa/inet.h>
#include "nlm.h"

static void	nlm_create_oh(netobj *obj, int pid);
static int	nlm_rqinit_1(lockd_req *rq);
static int	nlm_rqsend_1(lockd_req *rq);
static int	nlm_getconflock_1(lockd_req *rq, struct lklm_holder *hop);

lockd_proto	proto_nlm1 = {
	nlm_rqinit_1, nlm_rqsend_1, nlm_getconflock_1
};

/*
 * Initialize request struct, protocol-specific part.
 */
static int
nlm_rqinit_1(lockd_req *rq)
{
	nlm_create_oh(&rq->oh, rq->pid);
	rq->cookie = rpcmsg_gencookie();
	return 1;
}

/*
 * Send an NLM request.
 */
static int
nlm_rqsend_1(lockd_req *rq)
{
	lockd_host *	host;
	struct in_addr	addr;
	int		cookie;
	netobj		owner_o, cookie_o;
	int		status;

	host = rq->host;
	addr = host->addr;

	nlm_create_oh(&owner_o, rq->pid);

	cookie = rq->cookie;
	cookie_o.netobj_data = (void *) &cookie;
	cookie_o.netobj_len  = sizeof(cookie);

	switch (rq->type) {
	case RQ_TEST: {
		nlm_testargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.exclusive = rq->exclusive;
		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_TEST_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_TEST_MSG,
				(xdrproc_t)xdr_nlm_testargs, &args);

		break;
		}

	case RQ_LOCK: {
		nlm_lockargs 	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.block = rq->block;
		args.exclusive = rq->exclusive;
		args.reclaim = rq->reclaim;
		args.state = 0;	/* we don't do NSM yet */

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_LOCK_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_LOCK_MSG,
				(xdrproc_t)xdr_nlm_lockargs, &args);

		break;
		}

	case RQ_CANCEL: {
		nlm_cancargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;
		args.exclusive = rq->exclusive;
		args.block = rq->block;

		log(L_DEBUG, "calling NLM_CANCEL_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_CANCEL_MSG,
				(xdrproc_t)xdr_nlm_cancargs, &args);

		break;
		}

	case RQ_UNLOCK: {
		nlm_unlockargs	args;

		netobj_init(&(args.alock.fh), &(rq->fh));
		netobj_init(&(args.alock.oh), &owner_o);
		netobj_init(&(args.cookie), &cookie_o);

		args.alock.caller_name = host_myname();
		args.alock.svid = rq->pid;
		args.alock.l_offset = rq->l_offset;
		args.alock.l_len = rq->l_len;

		log(L_DEBUG, "calling NLM_UNLOCK_MSG on %s", inet_ntoa(addr));
		status = rpcmsg_send(addr, nlm_svcpg, NLM_VERS, NLM_UNLOCK_MSG,
				(xdrproc_t)xdr_nlm_unlockargs, &args);

		break;
		}
	}

	return status;
}

static int
nlm_getconflock_1(lockd_req *rq, struct lklm_holder *hop)
{
	struct nlm_testres	*resp;
	struct nlm_holder	*h;

	if (rq != RQ_TEST || !rq->result)
		return 0;
	resp = (struct nlm_testres *) rq->result;
	h = &(resp->stat.nlm_testrply_u.holder);

	strcpy(hop->alock.server_name, "<foobar>");
	hop->alock.fh.len = 0;
	hop->alock.pid = h->svid;
	hop->alock.l_offset = h->l_offset;
	hop->alock.l_len = h->l_len;
	hop->exclusive = h->exclusive;
	return 1;
}

/*
 * Create an owner handle. We don't bother with allocating memory and
 * such because these handles will be used only as an argument to 
 * rpcmsg_send, and are not kept around.
 */
#define MAXCOOKIEHOST	(MAX_NETOBJ_SZ - sizeof(int))
static void
nlm_create_oh(netobj *obj, int pid)
{
	static char	ownerhandle[MAX_NETOBJ_SZ];
	static int	ownerpidoff = 0;

	/* Create the owner handle. This is the concatenation of 
	 * our hostname and the caller's pid.
	 */
	if (ownerpidoff == 0) {
		strncpy(ownerhandle, host_myname(), MAXCOOKIEHOST);
		ownerhandle[MAXCOOKIEHOST] = '\0';
		if ((ownerpidoff = strlen(ownerhandle)) == 0) {
			log(L_ERROR, "can't get my own hostname");
			strcpy(ownerhandle, "foobar");
			ownerpidoff = 6;
		}
	}
	memcpy(ownerhandle + ownerpidoff, &pid, sizeof(int));
	obj->netobj_data = ownerhandle;
	obj->netobj_len  = ownerpidoff+sizeof(int);
}
