/*
 * grant.c	This module handles the granting of formerly blocked
 *		lock requests. Although this module uses methods
 *		that are very similar to those used in LKLM request
 *		handling, this is implemented separately, because the
 *		two tasks differ in quite a number of minor aspects.
 *
 * Copyright (C) Olaf Kirch, <okir@monad.swb.de>
 */

#include <signal.h>
#include "nlm.h"

/* Forward declaration. */
typedef struct grant_req	grant_req;

/*
 * Local function prototypes.
 */
static int		grt_xmit(grant_req * rq);
static void		grt_xmit_timeout(int timerid, void *clnt_data);
static grant_req *	grt_alloc(lockd_lock *lock);
static void		grt_unlink(grant_req * rq);

/*
 * This is the structure that holds all information for locks
 * in the process of being granted.
 */
struct grant_req {
	struct grant_req *	next;
	lockd_lock *		lock;
	int			cookie;

	struct timeval		wait;
	int			timerid;
	int			retries;
};

static grant_req *		requests = NULL;


/*
 * Queue and send a grant.
 */
int
nlm_grant(lockd_lock *lock)
{
	grant_req *	rq;
	int		status;

	rq = grt_alloc(lock);
	status = grt_xmit(rq);

	return status;
}

/*
 * When the lock is deleted for some reason, remove any pending
 * GRANTs first.
 */
void
nlm_dequeue_grantl(lockd_lock *lock)
{
	grant_req **	rqp;
	grant_req *	rq;

	for(rqp = &requests; *rqp != NULL; rqp = &((*rqp)->next)) {
		if ((rq = *rqp)->lock == lock)
			break;
	}
	if (*rqp != NULL) {
		*rqp = rq->next;
		rq->next = NULL;
	}
	xtimer_stop(rq->timerid);

	xfree (rq);
}

/*
 * Process result sent by client lockd.
 */
lockd_lock *
nlm_dequeue_grant(int cookie)
{
	lockd_lock *	lock;
	grant_req **	rqp;
	grant_req *	rq;

	for (rqp = &requests; *rqp != NULL; rqp = &((*rqp)->next)) {
		if ((rq = *rqp)->cookie == cookie)
			break;
	}
	if (*rqp == NULL)
		return NULL;

	xtimer_stop(rq->timerid);
	*rqp = rq->next;
	rq->next = NULL;

	lock = rq->lock;
	xfree (rq);

	return lock;
}

/*
 * Transmit a grant request.
 */
static int
grt_xmit(grant_req * rq)
{
	lockd_lock *	lock = rq->lock;
	nlm_testargs	args;
	int		cookie = rq->cookie;

	netobj_init(&(args.alock.fh), &(lock->file.back->fh));
	netobj_init(&(args.alock.oh), &(lock->oh));
	netobj_set(&(args.cookie), &cookie, sizeof(cookie));

	args.exclusive = (lock->mode == LCKMODE_EXCL)? 1 : 0;
	args.alock.caller_name = lock->host.back->name;
	args.alock.svid = lock->pid;
	args.alock.l_offset = lock->l_start;
	args.alock.l_len = lock->l_length;

	rpcmsg_send(lock->host.back->addr,
			nlm_cltpg, NLM_VERS, NLM_GRANTED_MSG,
			(xdrproc_t) xdr_nlm_testargs, &args);

	/* Set up the timer. */
	rq->timerid = xtimer_run(rq->wait, grt_xmit_timeout, rq);

	/* Tentatively: we believe it worked */
	return GRANT_INPROGRESS;
}

static void
grt_xmit_timeout(int timerid, void *clnt_data)
{
	grant_req *	rq = (grant_req *) clnt_data;

	if (rq->retries++ <= 6) {
		grt_xmit(rq);
		return;
	}
	grt_unlink(rq);
	lock_grant_status(GRANT_FAILED, rq->lock);

	xfree(rq);
}

/*
 * Allocate storage for a grant request.
 */
static grant_req *
grt_alloc(lockd_lock *lock)
{
	grant_req *	rq;

	rq = (grant_req *) xmalloc(sizeof(*rq));
	rq->next = requests;
	requests = rq;

	rq->cookie = rpcmsg_gencookie();
	rq->lock = lock;
	rq->retries = 0;
	rq->timerid = 0;
	timeval_set(rq->wait, 5, 0);

	return rq;
}

/*
 * Remove grant request from queue.
 */
static void
grt_unlink(grant_req * rq)
{
	grant_req **	rqp;

	for (rqp = &requests; *rqp != NULL; rqp = &((*rqp)->next)) {
		if (*rqp == rq) {
			(*rqp)->next = rq->next;
			break;
		}
	}
	rq->next = NULL;
}
