/*
 * nlm_mon.c	Handling of monitored locks.
 *
 * Copyright (C) 1995, Olaf Kirch <okir@monad.swb.de>
 */

#include "lock.h"
#include "host.h"
#include "nlm.h"
#include "sm_inter.h"
#include "log.h"
#include "misc.h"

static int	 nlm_reclaim_status(int status, lockd_req *rq);
static lockd_req *nlm_makereq(lockd_lock *lk, lockd_proto *prot);

/*
 * Monitor a host
 */
void
nlm_monitor(lockd_host *host, int state)
{
	if (!host->monitored) {
		log(L_DEBUG, "monitor host %s", host->name);
		/* not implemented yet */
	}
	host->monitored = 1;
	if (state != -1)
		host->state = state;
}

/*
 * Unmonitor a host. This should really be delayed so that subsequent
 * lock/unlock requests don't incur the overhead of talking to statd
 * each time.
 */
void
nlm_unmonitor(lockd_host *host)
{
	if (host->r_granted || host->r_blocked || host->l_reclaim ||
	    host->l_granted || host->l_blocked || host->l_waiting)
		return;
	if (host->monitored) {
		log(L_DEBUG, "unmonitor host %s", host->name);
		/* not implemented yet */
	}
	host->monitored = 0;
}

/*
 * Reclaim all locks. This affects locks both granted and blocked.
 * Previously granted locks are rescheduled with reclaim == 1; blocked
 * locks are sent with reclaim == 0. This will make the server reject
 * them during grace period, and nicely re-block them afterwards.
 * Requests that are currently being processed are not affected, because
 * they will be retransmitted again and again until grace period is over.
 *
 * We don't transmit all requests at once, but rather in 1msec intervals
 * in order to avoid congestion. Previously blocked locks are sent
 * with an overall delay of 5sec.
 * A 10msec timer resolution such as Linux's may hamper this scheme, but it
 * won't kill it.
 * FIXME: We should add a check to make sure a client holding
 * many locks on the server doesn't delay sending reclaims beyond the grace
 * period.
 */
#define STEP	1000
#define SECOND	1000000
void
nlm_reclaim_all(lockd_host *host, int newstate)
{
	struct timeval	wait = { 0, 0 };
	lockd_lock	*lk;
	lockd_req	*rq;

	host->monitored = 0;
	nlm_monitor(host, newstate);

	while ((lk = host->l_granted) != NULL) {
		rq = nlm_makereq(lk, &proto_nlm1);
		rq->reclaim = 1;
		rq->wait = wait;
		req_queue(rq, nlm_reclaim_status);

		if ((wait.tv_usec += STEP) >= SECOND) {
			wait.tv_usec -= SECOND;
			wait.tv_sec++;
		}
		lk_free(lk);
	}

	wait.tv_sec = 5; wait.tv_usec = 0;
	while ((rq = host->l_blocked) != NULL) {
		rq->wait = wait;
		req_retry(rq);
		if ((wait.tv_usec += STEP) >= SECOND) {
			wait.tv_usec -= SECOND;
			wait.tv_sec++;
		}
	}
}

static int
nlm_reclaim_status(int status, lockd_req *rq)
{
	if (status == RQ_GRANTED) {
		nlm_update_local(AC_LOCK, rq);
		return 1;
	}

	/* Bad things happening. Complain and send lock holder
	 * a SIGLOST. Not that it would help a lot unless the
	 * server turns off NFS during grace period.
	 */
	log(L_WARNING, "failed to reclaim a lock, status %d",
			status);
	kill(SIGLOST, rq->pid);

	if (status != RQ_BLOCKED)
		return 1;

	/* The lock-to-be-reclaimed was blocked. Naughty--cancel it
	 * now.
	 */
#ifdef notyet
	rq->type = RQ_CANCEL;
	rq->reclaim = 0;
	req_queue(rq, nlm_recancel_status);

	return 0;
#else
	return 1;
#endif
}

static lockd_req *
nlm_makereq(lockd_lock *lk, lockd_proto *prot)
{
	lockd_req	*rq;

	rq = req_alloc(lk->host.back, prot, RQ_LOCK, lk->pid);

	netobj_copy(&rq->fh, &lk->file.back->fh);
	rq->l_offset	= lk->l_start;
	rq->l_len	= lk->l_length;
	rq->exclusive	= lk->exclusive;
	rq->block	= lk->block;

	return rq;
}
