/*
 * request.c	This file handles lock requests we send to another lockd.
 *
 * These functions do almost the entire request processing and book-keeping.
 *
 * Copyright (C) 1995, Olaf Kirch  <okir@monad.swb.de>
 */

#include <signal.h>
#include "nlm.h"

static int		req_xmit(lockd_req *rq);
static void		req_xmit_timeout(int timerid, void * clnt_data);
static void		req_retry_timeout(int timerid, void * clnt_data);
#ifdef DEBUG
static void		req_verify(void);
#else
#define req_verify()	/* NOP */
#endif

static lockd_req *	request_retries = NULL;
static lockd_req *	request_process = NULL;

/*
 * Queue and send a request for TEST, LOCK, UNLOCK or CANCEL.
 * When calling this function, you must first lock the request Q.
 */
int
req_queue(lockd_req *rq, rqstatus_fn_t callback)
{
	int		type = rq->type;

	log(L_DEBUG, "queued %s request %d for %s", 
			(type == RQ_TEST)? "test" : (
			(type == RQ_LOCK)? "lock" : (
			(type == RQ_UNLOCK)? "unlock" : (
			(type == RQ_CANCEL)? "cancel" : "unknown"))),
			rq->cookie, rq->host->name);

	rq->callback = callback;
	req_insert(&(rq->host->l_waiting), rq);
	rq->nextreq = request_process;
	request_process = rq;
	req_verify();

	if (!rq->wait.tv_sec && !rq->wait.tv_usec)
		return (req_xmit(rq) == RPC_SUCCESS);
	req_retry(rq);
	return 1;
}

/*
 * A valid callback has been received from the server. Dequeue the
 * request and invoke the callback.
 */
void
req_process_result(lockd_req *rq, int status)
{
	log(L_DEBUG, "received callback for request %d", rq->cookie);

	req_unlink(rq);
	xtimer_stop(rq->timerid);
	if (rq->callback(status, rq)) {
		req_free(rq);
	} else {
		rq->state = LCKSTATE_PROCESSED;
	}
	req_verify();
}

/*
 * Send a request.
 * This function is called with all signals turned off.
 */
static int
req_xmit(lockd_req *rq)
{
	struct timeval	timeout;
	int		status;

	log(L_DEBUG, "transmit request %d (retry %d)",
				 rq->cookie, rq->retries);
	if ((status = rq->proto->send(rq)) != RQ_QUEUED) {
		req_unlink(rq);
		if (rq->callback(status, rq)) {
			req_free(rq);
		}
	} else {
		timeval_set(timeout, 5, 0);
		rq->timerid = xtimer_run(timeout, req_xmit_timeout, rq);
		rq->state = LCKSTATE_REQSENT;
	}
	req_verify();
	return status;
}

/*
 * A send timeout has occurred.
 * This function is called with all signals turned off.
 */
static void
req_xmit_timeout(int timerid, void * clnt_data)
{
	struct lockd_req * rq = (lockd_req *)clnt_data;

	if (rq->retries++ <= 6) {
		req_xmit(rq);
		return;
	}
	req_unlink(rq);
	if (rq->callback(RPC_TIMEDOUT, rq)) {
		req_free(rq);
	}
}

/*
 * Abort a queued request.
 */
void
req_abort(lockd_req * rq)
{
	xtimer_stop(rq->timerid);
}

/*
 * RPC status callback.
 */
void
req_status(int status, lockd_req *rq)
{
	log(L_DEBUG, "status callback for req %d, stat = %d",
				rq->cookie, status);

	req_unlink(rq);
	if (rq->callback(status, rq)) {
		/* Unlink and free request */
		req_free(rq);
	} else {
		if (rq->nextreq != NULL) {
			log(L_ERROR, "bad processed request in req_status");
		} else {
			rq->nextreq = request_process;
			request_process = rq;
		}
		rq->state = LCKSTATE_PROCESSED;
		req_verify();
	}
}

/*
 * Queue a request to be retried later.
 */
void
req_retry(lockd_req *rq)
{
	rq->timerid = xtimer_run(rq->wait, req_retry_timeout, rq);
}

/*
 * Retransmit the request now.
 */
static void
req_retry_timeout(int timerid, void * clnt_data)
{
	lockd_req *	rq;
	int		status;

	rq = (lockd_req *) clnt_data;
	if ((status = req_xmit(rq)) != RPC_SUCCESS) {
		log(L_WARNING, "failed to re-transmit request");
		req_status(status, rq);
	} else {
		req_insert(&(rq->host->l_waiting), rq);
	}
}

/*
 * Match a request that has seen a reply but no callback
 */
lockd_req *
req_match(int cookie)
{
	lockd_req **	rqp;
	lockd_req *	rq;

	for (rqp = &request_process; *rqp != NULL; rqp = &((*rqp)->nextreq)) {
		if ((rq = *rqp)->cookie == cookie) {
			*rqp = rq->nextreq;
			rq->nextreq = NULL;
			return rq;
		}
	}
	return NULL;
}

/*
 * Allocate a request.
 */
lockd_req *
req_alloc(lockd_host *host, lockd_proto *prot, int type, int pid)
{
	lockd_req *	rq;

	rq = (lockd_req *) xmalloc(sizeof(*rq));
	memset(rq, 0, sizeof(*rq));

	rq->host	= host;
	rq->next	= NULL;
	rq->prev	= NULL;
	rq->nextreq	= NULL;

	rq->proto	= prot;
	rq->type	= type;
	rq->pid		= pid;
	rq->callback	= NULL;
	rq->state	= LCKSTATE_INITIAL;
	rq->result	= NULL;

	timeval_zero(rq->wait);
	rq->retries	= 0;

	prot->init(rq);

	return rq;
/*
	netobj_copy(&(rq->fh), &(lkreq->fh));
	rq->l_offset	= lkreq->l_offset;
	rq->l_len	= lkreq->l_len;
	rq->pid		= lkreq->pid;
	rq->exclusive	= exclusive;
	rq->block	= block;
	rq->reclaim	= 0;

	rq->caller	= *svc_getcaller(rqstp->rq_xprt);
	rq->sockfd	= rqstp->rq_xprt->xp_sock;
	rq->xid		= svcudp_getxid(rqstp->rq_xprt);
	rq->callback	= callback;
	rq->state	= LCKSTATE_INITIAL;
	rq->cookie	= rpcmsg_gencookie();
	rq->result	= NULL;
	rq->type	= type;

	timeval_zero(rq->wait);
	rq->retries	= 0;

	return rq;
*/
}

/*
 * Free a processed request.
 */
void
req_free(lockd_req *rq)
{
	log(L_DEBUG, "req_free request %d", rq->cookie);
	if (request_process != NULL) {
		lockd_req **	rqp;

		rqp = &request_process;
		for (; *rqp != NULL; rqp = &((*rqp)->nextreq)) {
			if (*rqp == rq) {
				*rqp = rq->nextreq;
				break;
			}
		}
		if (*rqp == NULL) {
			log(L_ERROR, "bad unprocessed request in req_free");
		}
		/*
		 */
		req_verify();
	}

	if (rq->prev != NULL || rq->next != NULL) {
		log(L_ERROR, "freed lock request still in use");
		return;	/* rather have a memory leak than core dumps */
	}
	netobj_free(&(rq->fh));
	xfree (rq);
}

/*
 * Find a request.
 */
lockd_req *
req_find(lockd_req *head, int type, struct lklm_lock *lkreq,
					int exclusive, int block)
{
	for (; head != NULL; head = head->next) {
		if (head->pid == lkreq->pid && head->type == type &&
		    head->fh.n_len == lkreq->fh.len &&
		    !memcmp(head->fh.n_bytes, lkreq->fh.data, lkreq->fh.len) &&
		    head->l_offset == lkreq->l_offset &&
		    head->l_len == lkreq->l_len &&
		    (exclusive == -1 || head->exclusive == exclusive) &&
		    (block == -1 || head->block == block)) {
			return head;
		}
	}
	return NULL;
}

/*
 * Insert request in a queue.
 */
void
req_insert(lockd_req **head, lockd_req *rq)
{
	lockd_req *	tmp;

	if ((tmp = *head) != NULL) {
		tmp->prev = rq;
	}
	rq->next = tmp;
	rq->prev = NULL;
	*head = rq;
}

/*
 * Remove request from a queue.
 */
void
req_unlink(lockd_req *rq)
{
	lockd_host *	host;
	lockd_req *	tmp;

	if ((tmp = rq->prev) != NULL) {
		tmp->next = rq->next;
	} else if ((host = rq->host) != NULL) {
		if (host->l_waiting == rq) {
			host->l_waiting = rq->next;
		} else if (host->l_blocked == rq) {
			host->l_blocked = rq->next;
		} else if (host->l_reclaim == rq) {
			host->l_reclaim = rq->next;
		} else if (request_process == rq) {
			request_process = rq->next;
		} else if (request_retries == rq) {
			request_retries = rq->next;
		} else {
			log(L_ERROR, "orphan NLM request detected");
		}
	}
	if ((tmp = rq->next) != NULL) {
		tmp->prev = rq->prev;
	}
	rq->next = rq->prev = NULL;
}

/*
 * Check the request list for consistency.
 */
static void
req_verify(void)
{
	lockd_req	*rq, *rq2;

	if ((rq = request_process) == NULL)
		return;
	rq2 = rq->nextreq;
	while (rq2 != NULL && rq2->nextreq != NULL) {
		if (rq == rq2) {
			log(L_ERROR, "Detected loop in request queue!");
			pause();
		}
		rq2 = rq2->nextreq->nextreq;
		rq  = rq->nextreq;
	}
}
