/*
 *	rpcmisc - This module implements various RPC functions.
 *
 *	Copyright (C) 1995, O. Kirch <okir@monad.swb.de>
 */

#include <rpc/rpc.h>
#include <arpa/inet.h>
#include <signal.h>
#include "nlm.h"

#undef USE_RPC_UDPA

/*
 * Local functions.
 */
static int	rpcmsg_sendreply_common(struct sockaddr_in * caller,
					int sockfd, struct rpc_msg *msg);
static void	rpcmsg_initreply(void);

/*
 * This structure defines an entry in the CLIENT handle cache.
 */
typedef struct client_ent {
	struct sockaddr_in	addr;		/* server address */
	int			prog, vers, async;
	CLIENT			*clnt;		/* RPC client handle */
	time_t			lru;
} client_ent;
#define MAX_CLIENT_CACHE	20

static int			cookie = 0;	/* request matching */
static int			reply_initialized = 0;
static XDR			reply_xdr;
static void *			reply_buffer;
static client_ent		client_cache[MAX_CLIENT_CACHE];
static int			client_initialized = 0;


/*
 * Generate a unique cookie for an asynchronous RPC request.
 */
int
rpcmsg_gencookie(void)
{
	return cookie++;
}

/*
 * Send a synchronous RPC request. This is fairly straightforward.
 */
int
rpccall_send(struct in_addr addr, int prog, int vers, int proc,
			xdrproc_t encode, void *argp,
			xdrproc_t decode, void *resp)
{
	struct timeval	timeout = { 25, 0 };
	CLIENT		*client;
	enum clnt_stat	stat;

	log(L_DEBUG, "calling %d.%d@%s", prog, proc, inet_ntoa(addr));

	if ((client = rpc_getudpclnt(addr, prog, vers, 0)) == NULL)
		return 0;
	stat = clnt_call(client, proc, encode, argp, decode, resp, timeout);
	return (stat == RPC_SUCCESS);
}

/*
 * Send an asynchronous RPC request.
 * We have to use our own private clnt_call mutation here. This is an awful
 * kludge, but I have no idea how else to safely pass the cookie value to
 * clntudpa_call.
 */
int
rpcmsg_send(struct in_addr addr, int prog, int vers, int proc, 
			xdrproc_t encode, void *argp)
{
	struct timeval	timeout = { 0, 0 };
	CLIENT		*client;

	log(L_DEBUG, "calling %d.%d@%s (async)", prog, proc, inet_ntoa(addr));

	if ((client = rpc_getudpclnt(addr, prog, vers, 0)) == NULL)
		return RPC_CANTSEND;
	clnt_call(client, proc, encode, argp, (xdrproc_t)xdr_void,
							NULL, timeout);
	return RPC_SUCCESS;
}

/*
 * Send an RPC reply message.
 */
int
rpcmsg_sendreply(struct sockaddr_in *caller, int sockfd, u_long xid, 
					xdrproc_t encode, void *resp)
{
	struct rpc_msg	rply;

	log(L_DEBUG, "rpcmsg_sendreply(%s, %ld)",
				inet_ntoa(caller->sin_addr), xid);
	rply.rm_xid = xid;
	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_stat = SUCCESS;
	rply.acpted_rply.ar_verf.oa_flavor = AUTH_NULL;
	rply.acpted_rply.ar_verf.oa_length = 0;
	rply.acpted_rply.ar_verf.oa_base = NULL;
	rply.acpted_rply.ar_results.where = resp;
	rply.acpted_rply.ar_results.proc = encode;
	return rpcmsg_sendreply_common(caller, sockfd, &rply);
}

/*
 * Send an RPC error reply.
 */
int
rpcmsg_failreply(struct sockaddr_in *caller, int sockfd, u_long xid, int status)
{
	struct rpc_msg	rply;

	log(L_DEBUG, "rpcmsg_failreply(%s, %ld, %d)",
				inet_ntoa(caller->sin_addr), xid, status);
	rply.rm_xid = xid;
	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_stat = status;
	rply.acpted_rply.ar_verf.oa_flavor = AUTH_NULL;
	rply.acpted_rply.ar_verf.oa_length = 0;
	rply.acpted_rply.ar_verf.oa_base = NULL;
	return rpcmsg_sendreply_common(caller, sockfd, &rply);
}



static int
rpcmsg_sendreply_common(struct sockaddr_in * caller, int sockfd,	
						struct rpc_msg *msg)
{
	XDR *		xdrs;
	int		ok = 0;

	if (!reply_initialized)
		rpcmsg_initreply();

	xdrs = &reply_xdr;
	xdrs->x_op = XDR_ENCODE;
	XDR_SETPOS(xdrs, 0);
	if (xdr_replymsg(xdrs, msg)) {
		int	slen = XDR_GETPOS(xdrs);

		if (sendto(sockfd, reply_buffer, slen, 0,
			(struct sockaddr *) caller, sizeof(*caller)) == slen) {
			ok = 1;
		} else {
			log(L_DEBUG, "sendreply failed: %s", strerror(errno));
		}
	} else {
		log(L_DEBUG, "couldn't XDR reply message");
	}

	return ok;
}

static void
rpcmsg_initreply(void)
{
	reply_buffer = xmalloc(UDPMSGSIZE);
	xdrmem_create(&reply_xdr, reply_buffer, UDPMSGSIZE, XDR_ENCODE);
	reply_initialized = 1;
}

void
rpc_killudpclnt(struct in_addr addr, int prog, int vers)
{
	client_ent	*cp;
	int		i;

	if (!client_initialized)
		return;

	for (i = 0; i < MAX_CLIENT_CACHE; i++) {
		cp = client_cache + i;
		if (cp->clnt == NULL)
			continue;
		if (cp->addr.sin_addr.s_addr == addr.s_addr &&
		    cp->prog == prog && cp->vers == vers) {
			clnt_destroy(cp->clnt);
			cp->clnt = NULL;
		}
	}
}

/*
 * Get a client from the client cache.
 */
CLIENT *
rpc_getudpclnt(struct in_addr addr, int prog, int vers, int async)
{
	client_ent		*cp;
	struct timeval		timeout = {-1, -1};
	int			i, unused = -1, oldest = -1;
	time_t			lru, now;
	int			sock;

#ifndef USE_RPC_UDPA
	async = 0;
#endif

	if (!client_initialized) {
		for (i = 0; i < MAX_CLIENT_CACHE; i++)
			client_cache[i].clnt = NULL;
		client_initialized = 1;
	}

	now = time(NULL);
	for (i = 0, lru = now; i < MAX_CLIENT_CACHE; i++) {
		cp = client_cache + i;
		if (cp->clnt == NULL) {
			if (unused < 0) unused = i;
			continue;
		}
		if (cp->addr.sin_addr.s_addr == addr.s_addr &&
		    cp->prog == prog && cp->vers == vers &&
		    cp->async == async) {
			cp->lru = now;
			return cp->clnt;
		}
		if (cp->lru < lru) {
			oldest = i;
			lru = cp->lru;
		}
	}

	if (unused < 0) {
		clnt_destroy(client_cache[oldest].clnt);
		unused = oldest;
	}

	log(L_DEBUG, "creating UDP client for %d@%s", prog, inet_ntoa(addr));

	sock = RPC_ANYSOCK;
	cp = client_cache + unused;
	cp->addr.sin_addr = addr;
	cp->addr.sin_port = 0;
	cp->addr.sin_family = AF_INET;
	cp->prog = prog;
	cp->vers = vers;
	cp->async = async;
	cp->lru  = now;
	if (async) {
		log(L_ERROR, "ASYNC UDP currently not supported");
		return NULL;
	} else {
		cp->clnt = clntudp_create(&(cp->addr), prog, vers, timeout,
						&sock);
	}
	if (cp->clnt == NULL) {
		log(L_ERROR, "couldn't create client for %d@%s%s",
					prog, inet_ntoa(addr),
					clnt_spcreateerror(""));
		return NULL;
	}
	timeout.tv_sec  = 5;
	timeout.tv_usec = 0;

	clnt_control(cp->clnt, CLSET_RETRY_TIMEOUT, &timeout);

	return cp->clnt;
}

