/*
 * timer.c	Support for multiple timers.
 *
 * Copyright (C) 1995, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/time.h>
#include <signal.h>
#include "misc.h"
#include "log.h"

typedef struct xtimer {
	struct xtimer *		next;		/* linked list */
	struct timeval		when;		/* When to fire */
	xtimer_fn_t		handler;	/* user-provied handler */
	void *			clnt_data;	/* user-provided extra */
	int			cookie;		/* timer id */
} xtimer;

/*
 * Global variables.
 */
static xtimer *		timers = NULL;		/* List of timers */
static xtimer *		freelist = NULL;
static int		cookie = 1;		/* Unique timer id */

static void		xtimer_setup(struct timeval now);
static void		xtimer_timeout(int sig);
static void		xtimer_free(xtimer *ti);

/*
 * Create a watchdog and kick the interval timer.
 */
int
xtimer_run(struct timeval wait, xtimer_fn_t handler, void * clnt_data)
{
	struct itimerval	cur, new;
	struct timeval		now;
	xtimer 			*ti, **tip;
	int			retcookie;

	if ((ti = freelist) != NULL) {
		freelist = ti->next;
	} else {
		ti = (xtimer *) xmalloc(sizeof(*ti));
	}
	ti->handler = handler;
	ti->clnt_data = clnt_data;
	ti->cookie = cookie++;
	if (cookie == 0) cookie++;

	gettimeofday(&now, NULL);
	ti->when = now;
	timeval_bump(ti->when, wait);

	/* Stop the interval timer briefly */
	timeval_zero(new.it_interval);
	timeval_zero(new.it_value);
	setitimer(ITIMER_REAL, &new, &cur);

	for (tip = &(timers); *tip != NULL; tip = &((*tip)->next)) {
		if (timeval_lesseq(ti->when, (*tip)->when))
			break;
	}

	ti->next = *tip;
	*tip = ti;

	if (timers == ti) {
		xtimer_setup(now);
	} else {
		setitimer(ITIMER_REAL, &cur, NULL);
	}

	retcookie = ti->cookie;

	return retcookie;
}

int
xtimer_alarm(time_t seconds, xtimer_fn_t handler, void * clnt_data)
{
	struct timeval	wait;

	wait.tv_sec = seconds;
	wait.tv_usec = 0;
	return xtimer_run(wait, handler, clnt_data);
}

/*
 * Stop a running timer.
 * We don't adjust the interval timer, because that would open a whole
 * can of worms in terms of race conditions. We leave that to the natural
 * course of events (i.e. xtimer_timeout).
 */
void
xtimer_stop(int cookie)
{
	xtimer			*ti, **tip;

	for (tip = &(timers); *tip != NULL; tip = &((*tip)->next)) {
		if ((*tip)->cookie == cookie)
			break;
	}

	if ((ti = *tip) != NULL) {
		*tip = ti->next;
		xtimer_free(ti);
	}
}

/*
 * A timeout occurred. Invoke the handlers of all timers that have
 * expired (if there are any).
 */
static void
xtimer_timeout(int sig)
{
	struct timeval	now;
	xtimer		**tip;
	int		done;

	do {
		gettimeofday(&now, NULL);
		done = 1;
		
		for (tip = &(timers); *tip != NULL; tip = &((*tip)->next)) {
			if (!timeval_lesseq((*tip)->when, now))
				break;
		}

		if (tip != &timers) {
			xtimer	*next, *expired = timers;

			timers = *tip;
			for (; expired != NULL; expired = next) {
				expired->handler(expired->cookie,
							expired->clnt_data);
				next = expired->next;
				xtimer_free(expired);
			}
			done = 0;
		}
	} while (!done);

	xtimer_setup(now);
}

/*
 * Prime the interval timer to fire when the next xtimer expires.
 * All signals are blocked (see rpc_svcrun).
 */
static void
xtimer_setup(struct timeval now)
{
	struct itimerval	cur;
	struct sigaction	act;
	xtimer			*ti;

	if ((ti = timers) == NULL)
		return;

	timeval_zero(cur.it_interval);
	cur.it_value.tv_sec = ti->when.tv_sec - now.tv_sec;
	if (ti->when.tv_usec < now.tv_usec) {
		cur.it_value.tv_usec = 1000000 - ti->when.tv_usec
						+ now.tv_usec;
		cur.it_value.tv_sec--;
	} else {
		cur.it_value.tv_usec = ti->when.tv_usec - now.tv_usec;
	}

	act.sa_handler = xtimer_timeout;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGVTALRM);
	sigaddset(&act.sa_mask, SIGALRM);
	sigaddset(&act.sa_mask, SIGIO);
	act.sa_flags = SA_RESTART;
	sigaction(SIGALRM, &act, NULL);

	setitimer(ITIMER_REAL, &cur, NULL);
}

static void
xtimer_free(xtimer *ti)
{
	ti->next = freelist;
	freelist = ti;
}
