/*
 * unfsfh.c This file treats file handles issued by unfsd. The code
 *      was culled from the unfsd source which lists the following
 *      authors and copyright:
 *
 * Authors: Mark A. Shand, May 1988
 *      Donald J. Becker <becker@super.org>
 *      Rick Sladkey <jrs@world.std.com>
 *      Patrick Sweeney <pjs@raster.Kodak.COM>
 *      Orest Zborowski <obz@raster.Kodak.COM>
 *      Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *      Olaf Kirch, <okir@monad.swb.de>
 *
 *      Copyright 1988 Mark A. Shand
 *      This software maybe be used for any purpose provided
 *      the above copyright notice is retained.  It is supplied
 *      as is, with no warranty expressed or implied.
 */

#include <sys/stat.h>
#include <dirent.h>
#include "misc.h"
#include "unfsfh.h"
#include "log.h"

#define hash_psi(psi) (((psi)^((psi)>>8)^((psi)>>16)^((psi)>>24)) & 0xff)

/*
 * INODES and DEVICES.  NFS assumes that each file within an NFS mounted
 * file-system has a unique inode number.  Thus to mount an entire file
 * hierarchy, as this server sets out to do, requires mapping from inode/devs
 * to pseudo-inode.  Furthermore mount points must be detected and so that
 *  pseudo-inode("name") == pseudo-inode(direntry("name/../name"))
 * One option is to force the directory entry inode to correspond to the
 * result of the stat call, but this involves stat'ing every directory entry
 * during a readdir.  Instead we force the stat call to corresopnd to the
 * directory entry inode (see inner_getattr).  Of course this technique
 * requires that the parent directory is readable.  If it is not the normal
 * stat call result is used.  There is no chance of conflict because the
 * directory can never be read.
 *
 * In theory unique pseudo-inodes cannot be guaranteed, since inode/dev
 * contains 48 bits of information which must be crammed into an inode
 * number constrained to 32 bits.  Fortunately inodes numbers tend to be
 * small (often < 64k, almost always < 512k)
 */
static int
pseudo_inode(unsigned long inode, unsigned short dev)
{
    register dmajor, dminor;

    /*
     * Assuming major and minor numbers are small integers,
     * gravitate bits of dmajor & dminor device number to
     * high-order bits of word, to avoid clash with real inode num.
     */
    /* reverse (byte-wise) */
    dmajor = ((dev & 0xf0f) << 4) | ((dev & 0xf0f0) >> 4);
    dmajor = ((dmajor & 0x3333) << 2) | ((dmajor & 0xcccc) >> 2);
    dmajor = ((dmajor & 0x5555) << 1) | ((dmajor & 0xaaaa) >> 1);

    /* spread low-16 -> 32 with 0's in even posn */
    dmajor = ((dmajor & 0xff00) << 8) | (dmajor & 0xff);
    dmajor = ((dmajor & 0xf000f0) << 4) | (dmajor & 0xf000f);
    dmajor = ((dmajor & 0xc0c0c0c) << 2) | (dmajor & 0x3030303);
    dmajor = ((dmajor & 0x22222222) << 1) | (dmajor & 0x11111111);
    dminor = (dmajor & 0x5555) << 15;
    dmajor = dmajor & 0x55550000;

    return ((dmajor | dminor) ^ inode);
}

char *
unfsfh_buildpath(struct nfs_fh *fhp)
{
    struct stat sbuf;
    svc_fh      *h = (svc_fh *) fhp;
    int         i;
    int         psi;
    char        pathbuf[PATH_MAX + NAME_MAX + 1];
    long        cookie_stack[HP_LEN + 1];
    char        *slash_stack[HP_LEN];

    if (stat("/", &sbuf) < 0)
        return (NULL);

    psi = pseudo_inode(sbuf.st_ino, sbuf.st_dev);
    if (h->hash_path[0] == 0) {
        if (psi != h->psi)
            return (NULL);
        return xstrdup("/");
    }
    /* else */
    if (hash_psi(psi) != h->hash_path[1])
        return (NULL);

    strcpy(pathbuf, "/");
    cookie_stack[2] = 0;

    for (i = 2; i <= h->hash_path[0] + 1; i++) {
        struct dirent   *dp;
        DIR             *dir;
        char            *p;

    backtrack:
        if (stat(pathbuf, &sbuf) >= 0 && (dir = opendir(pathbuf)) != NULL) {
	    /* log(L_DEBUG, "trying %s...", pathbuf); */
            if (cookie_stack[i] != 0)
                seekdir(dir, cookie_stack[i]);
            while ((dp = readdir(dir))) {
                p = dp->d_name;
                if (p[0] == '.' && 
                   (p[1] == '\0' || (p[1] == '.' && p[2] == '\0')))
                    continue;

                psi = pseudo_inode(dp->d_ino, sbuf.st_dev);
		/* log(L_DEBUG, "\t%-10s psi %8d", dp->d_name, psi); */
                if (i == h->hash_path[0] + 1) {
                    if (psi == h->psi) {
                        /*GOT IT*/
                        strcat(pathbuf, p);
                        closedir(dir);
                        return xstrdup(pathbuf);
                    }
                } else {
                    if (hash_psi(psi) == h->hash_path[i]) {
                        /*PERHAPS WE'VE GOT IT */
                        cookie_stack[i] = telldir(dir);
                        cookie_stack[i + 1] = 0;
                        slash_stack[i] = pathbuf + strlen(pathbuf);
                        strcpy(slash_stack[i], p);
                        strcat(pathbuf, "/");

                        closedir(dir);
                        goto deeper;
                    }
                }
            }
            closedir(dir);
        }
        /* shallower */
        i--;
        if (i < 2)
            return NULL;  /* SEARCH EXHAUSTED */

        /* Prune path */
        *(slash_stack[i]) = '\0';
        goto backtrack;
    deeper:
        ;
    }
    return NULL;      /* actually not reached */
}

/*
 * The following is debug code that generates unfs file handles.
 */
#ifdef DEBUG
static int
path_psi(char *path)
{
    struct stat stb;

    if (lstat(path, &stb) < 0)
        return 0;
    if (S_ISDIR(stb.st_mode) && strcmp(path, "/") != 0) {
        /* Special case for directories--test for mount point. */
        struct stat ddbuf;
        char *sindx;
        char *fname;
        char squirrel;

        /* Find start of last component of path. */
        if ((sindx = strrchr(path, '/')) == path) {
            fname = ++sindx;
        } else
            fname = sindx + 1;

        /* Remove last element of path. */
        squirrel = *sindx;
        *sindx = '\0';
        if (lstat(path, &ddbuf) < 0) {
            *sindx = squirrel;
            return (0);
        }
        /* Sindx now points to directory entry name. */
        if (ddbuf.st_dev != stb.st_dev) {
            /* Directory is a mount point. */
            DIR *dirp;
            struct dirent *dp;

            if ((dirp = opendir(path)) == NULL) {
                *sindx = squirrel;  /* restore path */
            } else {
                *sindx = squirrel;  /* restore path */
                do {
                    if ((dp = readdir(dirp)) == NULL) {
                        closedir(dirp);
                        return (0);
                    }
                } while (strcmp(fname, dp->d_name) != 0);
                stb.st_dev = ddbuf.st_dev;
                stb.st_ino = dp->d_ino;
                closedir(dirp);
            }
        } else
            *sindx = squirrel;  /* restore path */
    }
    return pseudo_inode(stb.st_ino, stb.st_dev);
}

int
unfsfh_buildfh(char *path, struct nfs_fh *fhp)
{
    svc_fh  *h = (svc_fh *) fhp;
    int	    psi;
    char    *s;

    memset(fhp, 0, sizeof (*fhp));
    h->hash_path[0] = 0;

    if ((psi = path_psi("/")) == 0)
        return 1;

    s = path;
    while ((s = strchr(s + 1, '/')) != NULL) {
        if (++(h->hash_path[0]) >= HP_LEN)
            return 0;
        h->hash_path[h->hash_path[0]] = hash_psi(psi);
        *s = '\0';
	psi = path_psi(path);
	/* fprintf(stderr, "path component %40s psi=%8d\n", path, psi); */
	*s = '/';
	if (!psi)
            return 0;
    }
    if (*(strrchr(path, '/') + 1) != '\0') {
        if (++(h->hash_path[0]) >= HP_LEN)
            return 0;
        h->hash_path[h->hash_path[0]] = hash_psi(psi);
        if ((psi = path_psi(path)) == 0)
            return 0;
    }
    h->psi = psi;
    return 1;
}
#endif
